import datetime
import time
import typing
import hashlib


import flask
import flask_login
from saml2 import BINDING_HTTP_ARTIFACT, BINDING_SOAP, BINDING_HTTP_REDIRECT, response as saml_response, saml, time_util, element_to_extension_element, SAMLError
from saml2.assertion import _authn_context_class_ref, Policy
from saml2.authn_context import PASSWORDPROTECTEDTRANSPORT
from saml2.client import Saml2Client
from saml2.config import SPConfig, IdPConfig
from saml2.metadata import create_metadata_string
from saml2.saml import NAMEID_FORMAT_TRANSIENT, NAME_FORMAT_BASIC
from saml2.server import Server
from saml2.s_utils import factory, rndbytes, UnknownSystemEntity
from saml2.entity import create_artifact, ArtifactResponse
from saml2.samlp import response_from_string

from ..components import Component, get_component_by_uuid, set_component_fed_login_available
from .. import errors
from ...utils import FlaskResponseT
from ...models import SAMLArtifacts, SAMLMetadata, SAMLMetadataType
from ... import db

error_logging = {
    "version": 1,
    "loggers": {
        "saml2": {
            "level": "ERROR"
        }
    }
}


def get_idp_metadata(component: typing.Optional[Component] = None) -> typing.Any:
    config = get_pysaml_idp_config()
    return create_metadata_string(None, config=config, sign=False)


def get_sp_metadata(component: Component) -> typing.Any:
    config = get_pysaml_sp_config(component)
    return create_metadata_string(None, config=config, sign=False)


def get_pysaml_sp_config(component: Component) -> SPConfig:
    metadata = SAMLMetadata.query.filter_by(component_id=component.id, type=SAMLMetadataType.IDENTITY_PROVIDER_METADATA).first()
    conf = SPConfig()
    conf.load({
        "entityid": f"sp-{flask.current_app.config['FEDERATION_UUID']}",
        "description": flask.current_app.config['FEDERATION_UUID'],
        "delete_tmpfiles": True,
        "debug": 0,
        "logging": error_logging,
        "service": {
            "sp": {
                "want_response_signed": False,
                "want_assertions_signed": False,
                "want_assertions_or_response_signed": False,
                "allow_unsolicited": False,
                "want_authn_requests_signed": False,
                "endpoints": {
                    "assertion_consumer_service": [
                        (flask.url_for("frontend.assertion_consumer_service", _external=True), BINDING_HTTP_REDIRECT),
                        (flask.url_for("frontend.assertion_consumer_service", _external=True), BINDING_HTTP_ARTIFACT),
                        (flask.url_for("frontend.assertion_consumer_service", _external=True), BINDING_SOAP)
                    ],
                },
                "name_id_format": [NAMEID_FORMAT_TRANSIENT],
                "name_id_format_allow_create": True
            },
        },
        "metadata": {"inline": [metadata.metadata_xml] if metadata else []}
    })
    return conf


def get_pysaml_idp_config() -> IdPConfig:
    metadata = {"inline": [saml_metadata.metadata_xml for saml_metadata in SAMLMetadata.query.filter_by(type=SAMLMetadataType.SERVICE_PROVIDER_METADATA).all()]}

    config = IdPConfig()
    config.load({
        "entityid": f"idp-{flask.current_app.config['FEDERATION_UUID']}",
        "description": "SampleDB Federation IdP",
        "delete_tmpfiles": True,
        "debug": 0,
        "logging": error_logging,
        "service": {
            "idp": {
                "sign_assertion": False,
                "sign_response": False,
                "want_authn_requests_signed": False,
                "encrypt_assertion": False,
                "name": "SampleDB Federation IdP",
                "endpoints": {
                    "single_sign_on_service": [
                        (flask.url_for("frontend.federated_login_verify", _external=True), BINDING_HTTP_REDIRECT),
                    ],
                    "artifact_resolution_service": [
                        (flask.url_for("frontend.artifact_resolution_service", _external=True), BINDING_SOAP),
                    ]
                },
                "policy": {
                    "default": {
                        "lifetime": {"minutes": 30},
                        "attribute_restrictions": None,
                        "name_form": [NAME_FORMAT_BASIC],
                    }
                },
                "name_id_format": NAMEID_FORMAT_TRANSIENT
            }
        },
        "metadata": metadata,
    })
    return config


def sp_login(component: Component, shared_device: bool) -> flask.Response:
    config = get_pysaml_sp_config(component=component)
    sp = Saml2Client(config)

    relay_state = "shared_device" if shared_device else ""

    try:
        binding, destination = sp.pick_binding(
            "single_sign_on_service", [BINDING_HTTP_REDIRECT], "idpsso", entity_id=f"idp-{component.uuid}"
        )
    except UnknownSystemEntity:
        raise errors.InvalidSAMLRequestError()

    req_id, req = sp.create_authn_request(
        destination, sign=False
    )

    flask.session['saml_req_id'] = req_id
    flask.session['saml_sso_component'] = component.uuid
    flask.session['using_shared_device'] = shared_device

    http_args = sp.apply_binding(binding, str(req), destination, response=False, relay_state=relay_state)
    return flask.Response(response=http_args['data'], headers=http_args['headers'], status=http_args['status'])


def process_login(args: dict[str, str]) -> typing.Optional[FlaskResponseT]:
    if "SAMLRequest" not in args:
        raise errors.InvalidSAMLRequestError()
    msg = args["SAMLRequest"]
    if msg and " " in msg:
        msg = msg.replace(" ", "+")

    config = get_pysaml_idp_config()
    IDP = Server(config=config)

    try:
        authnRequest = IDP.parse_authn_request(msg, binding=BINDING_HTTP_REDIRECT)
    except saml_response.StatusError:
        raise errors.AuthnRequestParsingError()

    authnRequestMsg = authnRequest.message
    try:
        response_args = IDP.response_args(authnRequestMsg, bindings=[BINDING_HTTP_REDIRECT])
    except UnknownSystemEntity:
        return None
    except SAMLError:
        raise errors.InvalidSAMLRequestError()

    entityid = authnRequestMsg.issuer.text.strip()

    return _create_artifact_response(config, IDP, entityid, response_args)


def resolve_artifact(SAMLart: str) -> dict[str, typing.Any]:
    component = get_component_by_uuid(flask.session['saml_sso_component'])
    client = Saml2Client(get_pysaml_sp_config(component))
    message = client.artifact2message(SAMLart, "idpsso", sign=False)
    artifact_response = client.parse_artifact_resolve_response(message.text)

    if artifact_response.in_response_to != flask.session["saml_req_id"]:
        raise errors.MismatchedResponseToRequestError()

    identity = {
        attrib.friendly_name: attrib.attribute_value[0].text
        for attrib in artifact_response.assertion[0].attribute_statement[0].attribute
    }

    return identity


def generate_and_store_artifact(IDP: Server, authn_response: str, endpoint_index: int) -> typing.Any:
    """
    Corresponds to the saml2.entity.Entity.use_artifact method from PySAML2 module with the
    difference that the artifact is stored in the database.
    """
    message_handle = hashlib.sha1(str(authn_response).encode("utf-8"))
    message_handle.update(rndbytes())
    mhd = message_handle.digest()
    saml_art = create_artifact(IDP.config.entityid, mhd, endpoint_index)
    artifact = SAMLArtifacts(artifact=saml_art, message=authn_response)
    db.session.add(artifact)
    db.session.commit()
    return saml_art


def _create_artifact_response(config: dict[str, typing.Any], IDP: Server, ent_id: str, response_args: dict[str, typing.Any]) -> FlaskResponseT:
    del response_args['binding']
    response_args['release_policy'] = Policy(restrictions={"default": {"name_form": saml.NAME_FORMAT_BASIC}})

    identity = {
        "uid": flask_login.current_user.id,
        "email": flask_login.current_user.email,
    }

    response = IDP.create_authn_response(
        identity=identity,
        userid=flask_login.current_user.id,
        authn_statement=factory(
            saml.AuthnStatement,
            authn_instant=time_util.instant(time_stamp=int(time.time())),
            session_not_on_or_after=time_util.in_a_while(minutes=30),
            authn_context=_authn_context_class_ref(PASSWORDPROTECTEDTRANSPORT, None)
        ),
        encrypt_assertion=False,
        name_id=_generate_name_id(IDP.config.entityid, ent_id, str(identity['email'])),
        binding=BINDING_HTTP_REDIRECT,
        **response_args
    )

    artifact = generate_and_store_artifact(IDP, authn_response=str(response), endpoint_index=1)

    destination = IDP.pick_binding(
        "assertion_consumer_service", [BINDING_HTTP_ARTIFACT], "spsso", entity_id=ent_id
    )[1]

    response_result = IDP.apply_binding(BINDING_HTTP_ARTIFACT, artifact, destination=destination)
    artifact_url = response_result['url']
    return flask.redirect(artifact_url)


def eval_artifact_resolve(soap_request: bytes) -> dict[str, typing.Any]:
    config = get_pysaml_idp_config()
    IDP = Server(config=config)

    artifact_resolve_request = IDP.parse_artifact_resolve(soap_request)
    rinfo = IDP.response_args(artifact_resolve_request, bindings=[BINDING_SOAP])
    response = IDP._status_response(ArtifactResponse, artifact_resolve_request.issuer, None, sign=False, **rinfo)
    artifact_message = SAMLArtifacts.query.filter_by(artifact=artifact_resolve_request.artifact.text).first()

    if artifact_message is None:
        raise errors.UnavailableArtifactError()

    authn_response = response_from_string(artifact_message.message)
    response.extension_elements = [element_to_extension_element(authn_response)]

    result: dict[str, typing.Any] = IDP.apply_binding(binding=BINDING_SOAP, msg_str=str(response), response=True)
    return result


def delete_old_artifacts() -> None:
    SAMLArtifacts.query.filter(SAMLArtifacts.since <= datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(minutes=5)).delete()
    db.session.commit()


def _generate_name_id(idp_ent_id: int, sp_ent_id: str, text: str) -> saml.NameID:
    return saml.NameID(name_qualifier=idp_ent_id, sp_name_qualifier=sp_ent_id, format=saml.NAMEID_FORMAT_EMAILADDRESS, text=text)


def update_metadata(component: Component, updates: dict[str, typing.Any]) -> None:
    metadata_idp = SAMLMetadata.query.filter_by(component_id=component.id, type=SAMLMetadataType.IDENTITY_PROVIDER_METADATA).first()
    metadata_sp = SAMLMetadata.query.filter_by(component_id=component.id, type=SAMLMetadataType.SERVICE_PROVIDER_METADATA).first()

    if updates:
        if metadata_idp:
            metadata_idp.metadata_xml = updates['idp']
            metadata_idp.since = datetime.datetime.now(datetime.timezone.utc)
        else:
            metadata_idp = SAMLMetadata(component_id=component.id, metadata_xml=updates['idp'], type=SAMLMetadataType.IDENTITY_PROVIDER_METADATA)
            db.session.add(metadata_idp)

        if metadata_sp:
            metadata_sp.metadata_xml = updates['sp']
            metadata_sp.since = datetime.datetime.now(datetime.timezone.utc)
        else:
            metadata_sp = SAMLMetadata(component_id=component.id, metadata_xml=updates['sp'], type=SAMLMetadataType.SERVICE_PROVIDER_METADATA)
            db.session.add(metadata_sp)
        db.session.commit()
        set_component_fed_login_available(component.id, updates['enabled'])
    else:
        set_component_fed_login_available(component.id, False)


def check_component_locally_ready(component: Component) -> bool:
    return len(SAMLMetadata.query.filter_by(component_id=component.id).all()) == 2



import dataclasses
import enum
import typing

from . import errors
from .utils import cache
from .. import db, models, logic
from ..models import actions, topics, instruments, locations, objects


@dataclasses.dataclass(frozen=True)
class Topic:
    """
    This class provides an immutable wrapper around models.topics.Topic.
    """
    id: int
    name: typing.Dict[str, str]
    description: typing.Dict[str, str]
    short_description: typing.Dict[str, str]
    show_on_frontpage: bool
    show_in_navbar: bool
    description_is_markdown: bool
    short_description_is_markdown: bool
    order_index: typing.Optional[int]

    @classmethod
    def from_database(
            cls,
            topic: models.Topic
    ) -> 'Topic':
        wrapped_topic = Topic(
            id=topic.id,
            name=topic.name,
            description=topic.description,
            short_description=topic.short_description,
            show_on_frontpage=topic.show_on_frontpage,
            show_in_navbar=topic.show_in_navbar,
            description_is_markdown=topic.description_is_markdown,
            short_description_is_markdown=topic.short_description_is_markdown,
            order_index=topic.order_index
        )
        return wrapped_topic

    def __repr__(self) -> str:
        return f"<{type(self).__name__}(id={self.id!r})>"


@cache
def check_topic_exists(
        topic_id: int
) -> None:
    """
    Check whether a topic with the given topic ID exists.

    :param topic_id: the ID of an existing topic
    :raise errors.TopicDoesNotExistError: when no topic with the given topic ID exists
    """
    if not db.session.query(db.exists().where(models.Topic.id == topic_id)).scalar():
        raise errors.TopicDoesNotExistError()


def get_topic(topic_id: int) -> Topic:
    """
    Returns the topic with the given topic ID.

    :param topic_id: the ID of an existing topic
    :return: the topic
    :raise errors.TopicDoesNotExistError: when no topic with the
        given topic ID exists
    """
    topic = models.Topic.query.filter_by(id=topic_id).first()
    if topic is None:
        raise errors.TopicDoesNotExistError()
    return Topic.from_database(topic)


def get_topics(filter_frontpage: bool = False, filter_navbar: bool = False) -> typing.List[Topic]:
    """
    Return the list of all existing topics.

    :return: the topics
    """
    query = models.Topic.query
    query = query.order_by(db.nulls_last(models.Topic.order_index), models.Topic.id)

    return [
        Topic.from_database(topic)
        for topic in query.all()
        if (topic.show_in_navbar or not filter_navbar) and (topic.show_on_frontpage or not filter_frontpage)
    ]


def create_topic(
        *,
        name: typing.Dict[str, str],
        description: typing.Dict[str, str],
        short_description: typing.Dict[str, str],
        show_on_frontpage: bool,
        show_in_navbar: bool,
        description_is_markdown: bool,
        short_description_is_markdown: bool
) -> Topic:
    """
    Create a new topic.

    :param name: the unique names of the topic in a dict. Keys are lang codes and values names.
    :param description: (possibly empty) descriptions for the topic in a dict.
        Keys are lang codes and values are descriptions
    :param short_description: (possibly empty) short descriptions for the topic in a dict.
        Keys are lang codes and values are short descriptions
    :param show_on_frontpage: whether this topic should be shown on the frontpage
    :param show_in_navbar: whether this topic should be shown in the navbar
    :param description_is_markdown: whether the description is markdown
    :param short_description_is_markdown: whether the short description is markdown
    :return: the created topic
    """
    topic = models.Topic(
        name=name,
        description=description,
        short_description=short_description,
        show_on_frontpage=show_on_frontpage,
        show_in_navbar=show_in_navbar,
        description_is_markdown=description_is_markdown,
        short_description_is_markdown=short_description_is_markdown,
        order_index=None
    )
    db.session.add(topic)
    db.session.commit()
    return Topic.from_database(topic)


def update_topic(
        *,
        topic_id: int,
        name: typing.Dict[str, str],
        description: typing.Dict[str, str],
        short_description: typing.Dict[str, str],
        show_on_frontpage: bool,
        show_in_navbar: bool,
        description_is_markdown: bool,
        short_description_is_markdown: bool
) -> Topic:
    """
    Update an existing topic.

    :param topic_id: the ID of an existing topic
    :param name: the unique names of the topic in a dict. Keys are lang codes and values names.
    :param description: (possibly empty) descriptions for the topic in a dict.
        Keys are lang codes and values are descriptions
    :param short_description: (possibly empty) short descriptions for the topic in a dict.
        Keys are lang codes and values are short descriptions
    :param show_on_frontpage: whether this topic should be shown on the frontpage
    :param show_in_navbar: whether this topic should be shown in the navbar
    :param description_is_markdown: whether the description is markdown
    :param short_description_is_markdown: whether the short description is markdown
    :return: the created topic
    :raise errors.TopicDoesNotExistError: when no topic with the given topic ID exists
    """
    topic = models.Topic.query.filter_by(id=topic_id).first()
    if topic is None:
        raise errors.TopicDoesNotExistError()
    topic.name = name
    topic.description = description
    topic.short_description = short_description
    topic.show_on_frontpage = show_on_frontpage
    topic.show_in_navbar = show_in_navbar
    topic.description_is_markdown = description_is_markdown
    topic.short_description_is_markdown = short_description_is_markdown
    db.session.add(topic)
    db.session.commit()
    return Topic.from_database(topic)


def set_topics_order(topic_id_list: typing.List[int]) -> None:
    """
    Sets the `order_index` for all topics in the index_list, therefore the order in the list is used.

    :param topic_id_list: list of topic ids
    """
    for i, topic_id in enumerate(topic_id_list):
        topic = models.Topic.query.filter_by(id=topic_id).first()
        if topic is not None:
            topic.order_index = i
    db.session.commit()


def add_topic_to_order(topic: Topic) -> None:
    """
    Insert a topic into the current sort order.

    :param topic: the topic to insert
    """
    if topic.order_index is not None:
        # topic already has a place in the current sort order
        return
    topics = [
        Topic.from_database(other_topic)
        for other_topic in models.Topic.query.filter(models.Topic.order_index != db.null()).order_by(models.Topic.order_index).all()
    ]
    sorted_topics = topics + [topic]
    if topics:
        # check if topics are listed in order of their english names
        english_names = [
            other_topic.name.get('en', '').lower()
            for other_topic in topics
        ]
        english_lexicographical_order = english_names == sorted(english_names)
        if english_lexicographical_order:
            sorted_topics = sorted(sorted_topics, key=lambda t: t.name.get('en', '').lower())
    # update order indices
    index_list = [
        other_topic.id
        for other_topic in sorted_topics
    ]
    set_topics_order(index_list)


def _set_topics(
        mutable_object: typing.Union[actions.Action, instruments.Instrument, locations.Location],
        topic_ids: typing.Sequence[int]
) -> None:
    set_topics = topics.Topic.query.filter(topics.Topic.id.in_(topic_ids)).all()
    if len(set_topics) != len(set(topic_ids)):
        raise errors.TopicDoesNotExistError()
    mutable_object.topics.clear()
    for topic in set_topics:
        mutable_object.topics.append(topic)
    db.session.add(mutable_object)
    db.session.commit()


def set_action_topics(
        action_id: int,
        topic_ids: typing.Sequence[int]
) -> None:
    action = actions.Action.query.filter_by(id=action_id).first()
    if action is None:
        raise errors.ActionDoesNotExistError()
    _set_topics(action, topic_ids)


def set_instrument_topics(
        instrument_id: int,
        topic_ids: typing.Sequence[int]
) -> None:
    instrument = instruments.Instrument.query.filter_by(id=instrument_id).first()
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    _set_topics(instrument, topic_ids)


def set_location_topics(
        location_id: int,
        topic_ids: typing.Sequence[int]
) -> None:
    location = locations.Location.query.filter_by(id=location_id).first()
    if location is None:
        raise errors.LocationDoesNotExistError()
    _set_topics(location, topic_ids)


class TopicSource(enum.IntEnum):
    ACTION = 1
    INSTRUMENT = 2
    LOCATION = 3


def get_topic_ids_by_object_ids(
        object_ids: typing.Sequence[int],
        user_id: typing.Optional[int] = None
) -> typing.Dict[int, typing.Dict[int, typing.Set[TopicSource]]]:
    action_topics_select = db.select(
        objects.Objects.current_table.c.object_id, actions.topic_action_association_table.c.topic_id, TopicSource.ACTION, actions.topic_action_association_table.c.action_id
    ).filter(
        actions.topic_action_association_table.c.action_id == objects.Objects.current_table.c.action_id
    )
    if object_ids is not None:
        action_topics_select = action_topics_select.filter(
            objects.Objects.current_table.c.object_id.in_(object_ids)
        )
    instrument_topics_select = db.select(
        objects.Objects.current_table.c.object_id, instruments.topic_instrument_association_table.c.topic_id, TopicSource.INSTRUMENT, instruments.topic_instrument_association_table.c.instrument_id
    ).filter(
        objects.Objects.current_table.c.action_id == actions.Action.id,
        actions.Action.instrument_id == instruments.topic_instrument_association_table.c.instrument_id
    )
    if object_ids is not None:
        instrument_topics_select = instrument_topics_select.filter(
            objects.Objects.current_table.c.object_id.in_(object_ids)
        )

    location_topics_select = db.select(
        locations.ObjectLocationAssignment.object_id, locations.topic_location_association_table.c.topic_id, TopicSource.LOCATION, locations.topic_location_association_table.c.location_id
    ).filter(
        locations.ObjectLocationAssignment.location_id == locations.topic_location_association_table.c.location_id
    )
    if object_ids is not None:
        location_topics_select = location_topics_select.filter(
            locations.ObjectLocationAssignment.object_id.in_(object_ids)
        )
    object_id_topic_id_tuples = db.session.execute(
        action_topics_select.union(
            instrument_topics_select,
            location_topics_select
        )
    ).fetchall()
    topic_ids_by_object_id: typing.Dict[int, typing.Dict[int, typing.Set[TopicSource]]] = {}
    for object_id in object_ids:
        topic_ids_by_object_id[object_id] = {}
    action_ids = set()
    location_ids = set()
    for _, _, topic_source, topic_source_id in object_id_topic_id_tuples:
        if topic_source == TopicSource.ACTION:
            action_ids.add(topic_source_id)
        elif topic_source == TopicSource.LOCATION:
            location_ids.add(topic_source_id)
    action_permissions = logic.action_permissions.get_user_permissions_for_multiple_actions(list(action_ids), user_id)
    location_permissions = logic.location_permissions.get_user_permissions_for_multiple_locations(list(location_ids), user_id)
    for object_id, topic_id, topic_source, topic_source_id in object_id_topic_id_tuples:
        if topic_source == TopicSource.ACTION and models.Permissions.READ not in action_permissions.get(topic_source_id, models.Permissions.NONE):
            continue
        if topic_source == TopicSource.LOCATION and models.Permissions.READ not in location_permissions.get(topic_source_id, models.Permissions.NONE):
            continue
        if topic_id not in topic_ids_by_object_id[object_id]:
            topic_ids_by_object_id[object_id][topic_id] = set()
        topic_ids_by_object_id[object_id][topic_id].add(topic_source)
    return topic_ids_by_object_id



# coding: utf-8
"""
Add fed_id and component_id columns to object_location_assignments table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('object_location_assignments', 'fed_id'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE object_location_assignments
            ADD fed_id INTEGER,
            ADD component_id INTEGER,
            ADD FOREIGN KEY(component_id) REFERENCES components(id),
            ADD CONSTRAINT object_location_assignments_fed_id_component_id_key UNIQUE(fed_id, component_id)
    """))
    return True



# coding: utf-8
"""
Script for updating an instrument in SampleDB.

Usage: sampledb update_instrument <instrument_id> <name> <description>
"""

import sys
import typing

from .. import create_app
from ..logic.instruments import check_instrument_exists
from ..logic.instrument_translations import set_instrument_translation
from ..logic.errors import InstrumentDoesNotExistError
from ..logic.languages import Language


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 3:
        print(__doc__)
        sys.exit(1)
    instrument_id_str, name, description = arguments
    try:
        instrument_id = int(instrument_id_str)
    except ValueError:
        print("Error: instrument_id must be an integer", file=sys.stderr)
        sys.exit(1)
    app = create_app()
    with app.app_context():
        if app.config['DISABLE_INSTRUMENTS']:
            print('Error: instruments are disabled', file=sys.stderr)
            sys.exit(1)
        try:
            check_instrument_exists(instrument_id)
        except InstrumentDoesNotExistError:
            print('Error: no instrument with this id exists', file=sys.stderr)
            sys.exit(1)
        set_instrument_translation(
            language_id=Language.ENGLISH,
            instrument_id=instrument_id,
            name=name,
            description=description
        )
        print("Success: the instrument has been updated in SampleDB")



# coding: utf-8
"""
Replace string columns with JSON columns in locations table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    column_info = db.session.execute(db.text("""
        SELECT column_name, data_type
        FROM information_schema.columns
        WHERE table_name = 'locations'
    """)).fetchall()
    if ('name', 'json') in column_info:
        return False

    # Perform migration
    existing_data = [
        location_data
        for location_data in db.session.execute(db.text("""
            SELECT id, name, description
            FROM locations
        """)).fetchall()
    ]

    db.session.execute(db.text("""
        ALTER TABLE locations
        DROP COLUMN name;
    """))
    db.session.execute(db.text("""
        ALTER TABLE locations
        DROP COLUMN description;
    """))
    db.session.execute(db.text("""
        ALTER TABLE locations
        ADD name JSON NOT NULL DEFAULT '{}'::json;
    """))
    db.session.execute(db.text("""
        ALTER TABLE locations
        ADD description JSON NOT NULL DEFAULT '{}'::json;
    """))
    for id, name, description in existing_data:
        location_data = {
            'id': id,
            'name': name,
            'description': description
        }
        db.session.execute(db.text("""
            UPDATE locations
            SET name = json_build_object('en', :name), description = json_build_object('en', :description)
            WHERE id = :id;
        """), params=location_data)
    return True



# coding: utf-8
"""
Script for creating a user of type OTHER in SampleDB.

Usage: sampledb create_other_user <name> <email>
"""

import os
import string
import sys
import typing

from .. import create_app
from ..logic.users import create_user
from ..logic.authentication import add_other_authentication, is_login_available
from ..models import UserType


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 2 or not all(arguments):
        print(__doc__)
        sys.exit(1)
    name, email = arguments
    if not all(c in string.ascii_lowercase + string.digits + '_' for c in name):
        print("Error: name may only contain lower case characters, digits and underscores", file=sys.stderr)
        sys.exit(1)
    if not is_login_available(name):
        print('Error: name is already being used', file=sys.stderr)
        sys.exit(1)
    if '@' not in email[1:-1]:
        print("Error: email must be a valid email address", file=sys.stderr)
        sys.exit(1)
    password = ''.join([('00' + hex(c)[2:])[-2:] for c in os.urandom(16)])
    print(f"Note: the user will receive the password '{password}'")
    app = create_app()
    with app.app_context():
        user = create_user(name, email, UserType.OTHER)
        add_other_authentication(user.id, name, password)
        print(f"Success: the user has been created in SampleDB (#{user.id})")



# coding: utf-8
"""
Logic module for management of action translations

Action translations contain all language-dependent attributes of an action
for a specific language.
"""

import dataclasses
import typing
from flask_babel import _

from .. import db
from . import errors, actions, languages
from .languages import Language
from .. import models


@dataclasses.dataclass(frozen=True)
class ActionTranslation:
    """
    This class provides an immutable wrapper around models.action_translations.ActionTranslation.
    """
    action_id: int
    language_id: int
    name: str
    description: str
    short_description: str
    _language_cache: typing.List[languages.Language] = dataclasses.field(default_factory=list, kw_only=True, repr=False, compare=False)

    @classmethod
    def from_database(cls, action_translation: models.ActionTranslation) -> 'ActionTranslation':
        return ActionTranslation(
            action_id=action_translation.action_id,
            language_id=action_translation.language_id,
            name=action_translation.name or '',
            description=action_translation.description or '',
            short_description=action_translation.short_description or ''
        )

    @property
    def language(self) -> Language:
        if not self._language_cache:
            self._language_cache.append(languages.get_language(self.language_id))
        return self._language_cache[0]


def set_action_translation(
        language_id: int,
        action_id: int,
        name: str,
        description: str = '',
        short_description: str = '',
) -> ActionTranslation:
    """
    Create or update an action translation.

    :param language_id: the ID of an existing language
    :param action_id: the ID of an existing action
    :param name: the name of the action
    :param description: a (possibly empty) description for the action
    :param short_description: the new (possibly empty) short description
    :return: the created action translation
    :raise errors.LanguageDoesNotExistError: if no language with the given ID
        exists
    :raise errors.ActionDoesNotExistError: if no action with the given ID
        exists
    """

    action_translation = models.ActionTranslation.query.filter_by(
        language_id=language_id,
        action_id=action_id
    ).first()
    if action_translation is None:
        actions.check_action_exists(action_id)
        languages.get_language(language_id)
        action_translation = models.ActionTranslation(
            language_id=language_id,
            action_id=action_id,
            name=name,
            description=description,
            short_description=short_description,
        )
    else:
        action_translation.name = name
        action_translation.description = description
        action_translation.short_description = short_description
    db.session.add(action_translation)
    db.session.commit()
    return ActionTranslation.from_database(action_translation)


def get_action_translations_for_action(
        action_id: int,
        use_fallback: bool = False
) -> typing.List[ActionTranslation]:
    """
    Return all action translations for the action with the given action ID.

    If no translations exist and use_fallback is True, a fallback english
    translation will be returned.

    :param action_id: the ID of an existing action
    :param use_fallback: whether a fallback translation may be returned
    :return: a list with all action translations
    """
    action_translations = models.ActionTranslation.query.filter_by(action_id=action_id).order_by(models.ActionTranslation.language_id).all()
    if not action_translations and use_fallback:
        return [
            ActionTranslation(
                action_id=action_id,
                language_id=Language.ENGLISH,
                name=_('Unnamed Action (#%(action_id)s)', action_id=action_id),
                description='',
                short_description=''
            )
        ]
    return [
        ActionTranslation.from_database(action_translation)
        for action_translation in action_translations
    ]


def get_action_translation_for_action_in_language(
        action_id: int,
        language_id: int,
        use_fallback: bool = False
) -> ActionTranslation:
    """
    Return a translation for the given action ID and language ID.

    If use_fallback is False, this will fail unless the requested translation
    exists. Otherwise, this will build a fallback translation, using english
    translation values if no values for the given language exist and using a
    default name if no english name exists.

    :param action_id: the ID of an existing action
    :param language_id: the ID of an existing language
    :param use_fallback: a fallback translation may be built
    :return: the requested action translation
    :raise errors.ActionTranslationDoesNotExistError: when the action
        translation does not exist and use_fallback was False
    """
    action_translation = models.ActionTranslation.query.filter_by(
        action_id=action_id,
        language_id=language_id
    ).first()
    if not use_fallback:
        if action_translation is None:
            raise errors.ActionTranslationDoesNotExistError()
        else:
            return ActionTranslation.from_database(action_translation)

    if language_id == Language.ENGLISH:
        english_translation = action_translation
    else:
        english_translation = models.ActionTranslation.query.filter_by(
            action_id=action_id,
            language_id=Language.ENGLISH
        ).first()

    name = _('Unnamed Action (#%(action_id)s)', action_id=action_id)
    if action_translation is not None and action_translation.name:
        name = action_translation.name
    elif english_translation is not None and english_translation.name:
        name = english_translation.name

    description = ''
    if action_translation is not None and action_translation.description:
        description = action_translation.description
    elif english_translation is not None and english_translation.description:
        description = english_translation.description

    short_description = ''
    if action_translation is not None and action_translation.short_description:
        short_description = action_translation.short_description
    elif english_translation is not None and english_translation.short_description:
        short_description = english_translation.short_description

    return ActionTranslation(
        action_id=action_id,
        language_id=language_id if action_translation is not None else Language.ENGLISH,
        name=name,
        description=description,
        short_description=short_description
    )


def delete_action_translation(
        language_id: int,
        action_id: int
) -> None:
    """
    Delete an action translation.

    :param language_id: the ID of an existing language
    :param action_id: the ID of an existing action
    :raise errors.ActionTranslationDoesNotExistError: when there is no action
        translation for the given IDs
    """
    action_translation = models.ActionTranslation.query.filter_by(
        language_id=language_id,
        action_id=action_id
    ).first()
    if action_translation is None:
        raise errors.ActionTranslationDoesNotExistError()

    db.session.delete(action_translation)
    db.session.commit()



# coding: utf-8
"""

"""

import decimal
import functools
import json
import os
import typing
import pint
import pint.compat
import pint.facets.nonmultiplicative.definitions

from . import errors

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


@functools.wraps(pint.compat.log)  # type: ignore
def _decimal_compatible_log(x: typing.Any) -> typing.Any:
    if isinstance(x, decimal.Decimal):
        return x.ln()
    return pint.compat.log(x)  # type: ignore


# monkey patch the log function used for non-multiplicative units in pint, as
# numpy.log (used by pint) expects the presence of a method log(), but the
# (natural) logarithm of a decimal.Decimal instance is calculated using ln().
pint.facets.nonmultiplicative.definitions.log = _decimal_compatible_log  # type: ignore

ureg = pint.UnitRegistry(non_int_type=decimal.Decimal)
ureg.load_definitions(os.path.join(os.path.dirname(__file__), 'unit_definitions.txt'))

int_ureg = pint.UnitRegistry()
int_ureg.load_definitions(os.path.join(os.path.dirname(__file__), 'unit_definitions.txt'))

with open(os.path.join(os.path.dirname(__file__), 'un_cefact_code_unit_pairs.json'), encoding='utf-8') as _json_file:
    _un_cefact_code_unit_pairs: typing.List[typing.Tuple[str, str]] = [
        (code, unit)
        for code, unit in json.load(_json_file)
    ]
_un_cefact_code_to_symbol = {
    code: unit
    for code, unit in _un_cefact_code_unit_pairs
}
_symbol_to_un_cefact_code = {
    unit: code
    for code, unit in _un_cefact_code_unit_pairs
}


def prettify_units(units: typing.Optional[typing.Union[str, pint._typing.UnitLike]]) -> str:
    """
    Returns a prettified version of the units, if defined, otherwise returns the units unaltered.
    :param units: The pint units or their string representation
    :return: The prettified units
    """
    if units is None:
        units = '1'
    units_str = str(units).strip()
    units_str = {
        'degC': '\xb0C',
        'degree_Celsius': '\xb0C',
        'deg': '\xb0',
        'degree': '\xb0',
        'percent': '%',
        '': '—',
        '1': '—'
    }.get(units_str, units_str)
    return units_str


def get_dimensionality_for_units(units: typing.Optional[typing.Union[str, pint._typing.UnitLike]]) -> str:
    """
    Return the dimensionality of given units.

    :param units: the units to get the dimensionality for
    :return: dimensionality string in pint format
    :raise errors.InvalidUnitsError: if the units cannot be understood
    """
    if units is None:
        units = '1'
    try:
        return str(int_ureg.Unit(units).dimensionality)
    except Exception:
        raise errors.InvalidUnitsError()


def get_magnitude_in_base_units(
        magnitude: decimal.Decimal,
        units: typing.Union[str, pint._typing.UnitLike]
) -> decimal.Decimal:
    """
    Convert a given magnitude in a given unit to base units.

    :param magnitude: the given magnitude
    :param units: the units to get the dimensionality for
    :return: magnitude in base units
    :raise errors.InvalidUnitsError: if the units cannot be understood
    """
    try:
        return typing.cast(decimal.Decimal, ureg.Quantity(magnitude, ureg.Unit(units)).to_base_units().magnitude)
    except Exception:
        raise errors.InvalidUnitsError()


@functools.cache
def get_un_cefact_code_for_unit(
        unit: str
) -> typing.Optional[str]:
    """
    Get the UN CEFACT code for a given unit.

    :param unit: the unit to get the UN CEFACT code for
    :return: the UN CEFACT code for the given unit, or None
    """
    code = _symbol_to_un_cefact_code.get(unit)
    if code is not None:
        return code
    try:
        pint_unit = ureg.Unit(unit)
    except Exception:
        return None
    for code, other_unit in _un_cefact_code_unit_pairs:
        try:
            pint_other_unit = ureg.Unit(other_unit)
        except Exception:
            continue
        if pint_other_unit == pint_unit:
            return code
    return None


@functools.cache
def get_unit_for_un_cefact_code(
        code: str
) -> typing.Optional[str]:
    """
    Get the unit for a given UN CEFACT code.

    :param code: the UN CEFACT code to get the unit for
    :return: the unit for the given UN CEFACT code, or None
    """
    unit = _un_cefact_code_to_symbol.get(code)
    if unit is None:
        return None
    try:
        ureg.Unit(unit)
    except Exception:
        return None
    return unit


@functools.cache
def get_old_dimensionality(units: typing.Optional[str]) -> str:
    """
    Get the dimensionality string in the style of pint 0.23 and earlier.

    :param units: the units string to get the dimensionality for
    :return: the dimensionality string
    """
    if units is None:
        return 'dimensionless'
    pint_units = int_ureg.Unit(units)
    dimensionality = pint_units.dimensionality
    numerator_items = []
    denominator_items = []
    for dimension, exponent in dimensionality.items():
        if exponent > 0:
            numerator_items.append((dimension, exponent))
        else:
            denominator_items.append((dimension, -exponent))
    # sort by dimension, rather than by exponent
    numerator_items.sort()
    denominator_items.sort()
    numerator_str = ' * '.join([
        f'{dimension} ** {int(exponent)}' if exponent > 1 else str(dimension)
        for dimension, exponent in numerator_items
    ])
    denominator_str = ' / '.join([
        f'{dimension} ** {int(exponent)}' if exponent > 1 else str(dimension)
        for dimension, exponent in denominator_items
    ])
    if denominator_str and numerator_str:
        return numerator_str + ' / ' + denominator_str
    elif numerator_str:
        return numerator_str
    elif denominator_str:
        return '1 / ' + denominator_str
    else:
        return 'dimensionless'



import flask
import flask_login

from ..utils import Resource, ResponseData
from ...logic import errors
from ...logic.actions import check_action_exists
from ...logic.instruments import check_instrument_exists
from ...logic.favorites import add_favorite_action, remove_favorite_action, get_user_favorite_action_ids, add_favorite_instrument, remove_favorite_instrument, get_user_favorite_instrument_ids


class FavoriteAction(Resource):
    @flask_login.login_required
    def put(self, action_id: int) -> ResponseData:
        if flask_login.current_user.is_readonly:
            return {
                "message": "user is readonly"
            }, 403
        try:
            check_action_exists(action_id)
        except errors.ActionDoesNotExistError:
            return {
                "message": f"action {action_id} does not exist"
            }, 404
        is_favorite = action_id in get_user_favorite_action_ids(user_id=flask_login.current_user.id)
        if is_favorite:
            return {
                "message": f"action {action_id} is already a favorite"
            }, 409
        add_favorite_action(action_id=action_id, user_id=flask_login.current_user.id)
        return {
            "message": f"action {action_id} has been added to favorite actions"
        }, 200

    @flask_login.login_required
    def delete(self, action_id: int) -> ResponseData:
        if flask_login.current_user.is_readonly:
            return {
                "message": "user is readonly"
            }, 403
        try:
            check_action_exists(action_id)
        except errors.ActionDoesNotExistError:
            return {
                "message": f"action {action_id} does not exist"
            }, 404
        is_favorite = action_id in get_user_favorite_action_ids(user_id=flask_login.current_user.id)
        if not is_favorite:
            return {
                "message": f"action {action_id} is not a favorite"
            }, 404
        remove_favorite_action(action_id=action_id, user_id=flask_login.current_user.id)
        return {
            "message": f"action {action_id} has been removed from favorite actions"
        }, 200


class FavoriteInstrument(Resource):
    @flask_login.login_required
    def put(self, instrument_id: int) -> ResponseData:
        if flask_login.current_user.is_readonly:
            return {
                "message": "user is readonly"
            }, 403
        if flask.current_app.config['DISABLE_INSTRUMENTS']:
            return {
                "message": "instruments are disabled"
            }, 404
        try:
            check_instrument_exists(instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        is_favorite = instrument_id in get_user_favorite_instrument_ids(user_id=flask_login.current_user.id)
        if is_favorite:
            return {
                "message": f"instrument {instrument_id} is already a favorite"
            }, 409
        add_favorite_instrument(instrument_id=instrument_id, user_id=flask_login.current_user.id)
        return {
            "message": f"instrument {instrument_id} has been added to favorite instruments"
        }, 200

    @flask_login.login_required
    def delete(self, instrument_id: int) -> ResponseData:
        if flask_login.current_user.is_readonly:
            return {
                "message": "user is readonly"
            }, 403
        if flask.current_app.config['DISABLE_INSTRUMENTS']:
            return {
                "message": "instruments are disabled"
            }, 404
        try:
            check_instrument_exists(instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        is_favorite = instrument_id in get_user_favorite_instrument_ids(user_id=flask_login.current_user.id)
        if not is_favorite:
            return {
                "message": f"instrument {instrument_id} is not a favorite"
            }, 404
        remove_favorite_instrument(instrument_id=instrument_id, user_id=flask_login.current_user.id)
        return {
            "message": f"instrument {instrument_id} has been removed from favorite instruments"
        }, 200



# coding: utf-8
"""

"""

import typing

import flask

from . import actions
from . import users
from .permissions import ResourcePermissions
from .actions import Action
from ..models import Permissions, UserActionPermissions, GroupActionPermissions, ProjectActionPermissions, AllUserActionPermissions
from ..models.instruments import instrument_user_association_table
from .. import db, models

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


action_permissions = ResourcePermissions(
    resource_id_name='action_id',
    all_user_permissions_table=AllUserActionPermissions,
    anonymous_user_permissions_table=None,
    user_permissions_table=UserActionPermissions,
    group_permissions_table=GroupActionPermissions,
    project_permissions_table=ProjectActionPermissions,
    check_resource_exists=lambda resource_id: actions.check_action_exists(action_id=resource_id)
)


def get_action_permissions_for_all_users(action_id: int) -> Permissions:
    return action_permissions.get_permissions_for_all_users(resource_id=action_id)


def set_action_permissions_for_all_users(action_id: int, permissions: Permissions) -> None:
    action_permissions.set_permissions_for_all_users(resource_id=action_id, permissions=permissions)


def get_action_permissions_for_users(action_id: int) -> typing.Dict[int, Permissions]:
    return action_permissions.get_permissions_for_users(resource_id=action_id)


def set_user_action_permissions(action_id: int, user_id: int, permissions: Permissions) -> None:
    action_permissions.set_permissions_for_user(resource_id=action_id, user_id=user_id, permissions=permissions)


def get_action_permissions_for_groups(action_id: int) -> typing.Dict[int, Permissions]:
    return action_permissions.get_permissions_for_groups(resource_id=action_id)


def set_group_action_permissions(action_id: int, group_id: int, permissions: Permissions) -> None:
    action_permissions.set_permissions_for_group(resource_id=action_id, group_id=group_id, permissions=permissions)


def get_action_permissions_for_projects(action_id: int) -> typing.Dict[int, Permissions]:
    return action_permissions.get_permissions_for_projects(resource_id=action_id)


def set_project_action_permissions(action_id: int, project_id: int, permissions: Permissions) -> None:
    action_permissions.set_permissions_for_project(resource_id=action_id, project_id=project_id, permissions=permissions)


def get_user_action_permissions(
        action_id: int,
        user_id: typing.Optional[int],
        *,
        include_groups: bool = True,
        include_projects: bool = True,
        include_admin_permissions: bool = True,
        include_instrument_responsible_users: bool = True
) -> Permissions:
    return get_user_permissions_for_multiple_actions(
        action_ids=[action_id],
        user_id=user_id,
        include_groups=include_groups,
        include_projects=include_projects,
        include_admin_permissions=include_admin_permissions,
        include_instrument_responsible_users=include_instrument_responsible_users
    )[action_id]


def get_user_permissions_for_multiple_actions(
        action_ids: typing.Sequence[int],
        user_id: typing.Optional[int],
        *,
        include_groups: bool = True,
        include_projects: bool = True,
        include_admin_permissions: bool = True,
        include_instrument_responsible_users: bool = True,
        max_permissions: Permissions = Permissions.GRANT
) -> typing.Mapping[int, Permissions]:
    if not action_ids:
        return {}

    if user_id is None:
        return {
            action_id: Permissions.NONE
            for action_id in action_ids
        }

    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        include_instrument_responsible_users = False

    additional_permissions = {
        action_id: Permissions.NONE
        for action_id in action_ids
    }

    for action_id in action_ids:
        # users have GRANT permissions for actions they own
        owner_id = actions.get_action_owner_id(action_id)
        if owner_id is not None and owner_id == user_id:
            additional_permissions[action_id] = Permissions.GRANT

    if include_instrument_responsible_users:
        # instrument responsible users have GRANT permissions for actions of their instrument
        user_is_instrument_responsible = _is_user_responsible_for_actions_instruments(
            user_id=user_id,
            action_ids=[
                action_id
                for action_id in action_ids
                if Permissions.GRANT not in additional_permissions[action_id]
            ]
        )
        for action_id in user_is_instrument_responsible:
            if user_is_instrument_responsible[action_id]:
                additional_permissions[action_id] = max(additional_permissions[action_id], Permissions.GRANT)

    # resource independent permissions
    return action_permissions.get_permissions_for_user_for_multiple_resources(
        resource_ids=action_ids,
        user_id=user_id,
        include_all_users=True,
        include_groups=include_groups,
        include_projects=include_projects,
        include_admin_permissions=include_admin_permissions,
        limit_readonly_users=True,
        additional_permissions=additional_permissions,
        max_permissions=max_permissions
    )


def _is_user_responsible_for_actions_instruments(
        user_id: int,
        action_ids: typing.Sequence[int]
) -> typing.Mapping[int, bool]:
    responsible_action_id_rows = db.session.query(models.Action.id).filter(
        models.Action.id.in_(tuple(action_ids)),
        models.Action.instrument_id == instrument_user_association_table.c.instrument_id,
        instrument_user_association_table.c.user_id == user_id
    ).all()
    return {
        action_id: (action_id,) in responsible_action_id_rows
        for action_id in action_ids
    }


def get_actions_with_permissions(user_id: typing.Optional[int], permissions: Permissions, action_type_id: typing.Optional[int] = None) -> typing.List[Action]:
    """
    Get all actions which a user has the given permissions for.

    Return an empty list if called with Permissions.NONE.

    :param user_id: the ID of an existing user, or None
    :param permissions: the minimum permissions required for the actions for
        the given user
    :param action_type_id: None or the ID of an existing action type
    :return: the actions with the given permissions for the given
        user
    :raise errors.UserDoesNotExistError: if no user with the given user ID
        exists
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    """
    if user_id is not None:
        # ensure that the user can be found
        users.check_user_exists(user_id)
    if permissions == Permissions.NONE:
        return []
    all_actions = actions.get_actions(action_type_id=action_type_id)
    action_permissions = get_user_permissions_for_multiple_actions(
        action_ids=[action.id for action in all_actions],
        user_id=user_id,
        max_permissions=permissions
    )
    actions_with_permissions = []
    for action in all_actions:
        if permissions in action_permissions.get(action.id, Permissions.NONE):
            actions_with_permissions.append(action)
    return actions_with_permissions


def get_sorted_actions_for_user(
        user_id: int,
        action_type_id: typing.Optional[int] = None,
        owner_id: typing.Optional[int] = None,
        include_hidden_actions: bool = False
) -> typing.List[Action]:
    """
    Get sorted actions for a user, optionally filtered by type and owning user.

    The actions are sorted by:
    - favorite / not favorite
    - instrument name (independent actions first)
    - action name

    :param user_id: the ID of the user to return actions for
    :param action_type_id: the ID of the actions' type, or None
    :param owner_id: the ID of the actions' owner, or None
    :param include_hidden_actions: whether hidden actions should be included
    :return: the sorted list of actions
    :raise errors.UserDoesNotExistError: if no user with the given user ID
        exists
    """
    user = users.get_user(user_id)
    possibly_visible_actions = []
    for action in actions.get_actions(
        action_type_id=action_type_id
    ):
        if owner_id is not None and action.user_id != owner_id:
            continue
        if not include_hidden_actions and action.is_hidden and not user.is_admin and owner_id != user_id:
            continue
        possibly_visible_actions.append(action)
    visible_actions = []
    action_permissions = {}
    permissions = get_user_permissions_for_multiple_actions(
        user_id=user_id,
        action_ids=[action.id for action in possibly_visible_actions],
        max_permissions=Permissions.READ
    )
    for action in possibly_visible_actions:
        if Permissions.READ in permissions[action.id]:
            visible_actions.append(action)
            action_permissions[action.id] = permissions[action.id]
    return actions.sort_actions_for_user(actions=visible_actions, user_id=user_id)



# coding: utf-8
"""
Add use_instrument_topics column to actions table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('actions', 'use_instrument_topics'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE actions
        ADD use_instrument_topics BOOLEAN DEFAULT FALSE NOT NULL
    """))
    db.session.execute(db.text("""
        UPDATE actions
        SET use_instrument_topics = TRUE
        WHERE instrument_id IS NOT NULL AND id NOT IN (SELECT action_id FROM action_topics)
    """))
    return True



# coding: utf-8
"""
Models for imported .eln files.
"""
import datetime
import typing

from sqlalchemy.orm import Mapped, Query, relationship

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .users import User


class ELNImport(Model):
    __tablename__ = 'eln_imports'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    file_name: Mapped[str] = db.Column(db.String, nullable=False)
    binary_data: Mapped[bytes] = db.deferred(db.Column(db.LargeBinary, nullable=False))
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id', use_alter=True, ondelete='CASCADE', name="fk_eln_imports_user_id"), nullable=False)
    upload_utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    import_utc_datetime: Mapped[typing.Optional[datetime.datetime]] = db.Column(db.TIMESTAMP(timezone=True), nullable=True)
    invalid_reason: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    user: Mapped['User'] = relationship('User', foreign_keys=[user_id])
    signed_by: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    objects: Mapped[typing.List['ELNImportObject']] = relationship('ELNImportObject', back_populates='eln_import')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ELNImport"]]


class ELNImportAction(Model):
    __tablename__ = 'eln_import_actions'

    action_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('actions.id'), primary_key=True)
    action_type_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('action_types.id'), nullable=True, unique=True)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ELNImportAction"]]


class ELNImportObject(Model):
    __tablename__ = 'eln_import_objects'

    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('objects_current.object_id'), primary_key=True)
    eln_import_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('eln_imports.id'), nullable=False)
    url: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    eln_import: Mapped[ELNImport] = relationship('ELNImport', back_populates='objects')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ELNImportObject"]]



# coding: utf-8
"""
Create location log entries for object location assignments created before the location log was implemented.
"""
import json
import typing

import flask_sqlalchemy

from ..location_log import LocationLogEntryType


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    location_log_entries_exist = db.session.execute(db.text("""
        SELECT id
        FROM location_log_entries
    """)).fetchone() is not None
    if location_log_entries_exist:
        return False

    object_location_assignments = db.session.execute(db.text("""
        SELECT id, location_id, user_id, object_id, utc_datetime
        FROM object_location_assignments
        ORDER BY utc_datetime ASC
    """)).fetchall()

    if not object_location_assignments:
        return False

    location_id_by_object_id: typing.Dict[int, int] = {}
    log_entries = []
    for object_location_assignment in object_location_assignments:
        object_location_assignment_id, location_id, user_id, object_id, utc_datetime = object_location_assignment
        previous_location_for_object = location_id_by_object_id.get(object_id)
        location_id_by_object_id[object_id] = location_id
        if location_id is not None and previous_location_for_object == location_id:
            log_entries.append({
                'type': LocationLogEntryType.CHANGE_OBJECT.name,
                'location_id': location_id,
                'utc_datetime': utc_datetime,
                'user_id': user_id,
                'data': json.dumps({
                    'object_location_assignment_id': object_location_assignment_id
                })
            })
        else:
            if previous_location_for_object is not None:
                log_entries.append({
                    'type': LocationLogEntryType.REMOVE_OBJECT.name,
                    'location_id': previous_location_for_object,
                    'utc_datetime': utc_datetime,
                    'user_id': user_id,
                    'data': json.dumps({
                        'object_location_assignment_id': object_location_assignment_id
                    })
                })
            if location_id is not None:
                log_entries.append({
                    'type': LocationLogEntryType.ADD_OBJECT.name,
                    'location_id': location_id,
                    'utc_datetime': utc_datetime,
                    'user_id': user_id,
                    'data': json.dumps({
                        'object_location_assignment_id': object_location_assignment_id
                    })
                })
    for log_entry in log_entries:
        db.session.execute(db.text("""
            INSERT INTO location_log_entries
            (type, location_id, utc_datetime, user_id, data)
            VALUES
            (CAST(:type AS locationlogentrytype), :location_id, :utc_datetime, :user_id, CAST(:data AS JSONB))
        """), params=log_entry)
    return True



# coding: utf-8
"""
Add EDIT_URL enum value to FileLogEntryType enum.
"""

import flask_sqlalchemy

from .utils import enum_value_migration


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    return enum_value_migration('filelogentrytype', 'EDIT_URL')



# coding: utf-8
"""

"""
import typing

import flask
import flask_login
from flask_wtf import FlaskForm
from wtforms.fields import BooleanField
from flask_babel import _

from . import frontend
from .. import logic
from .utils import check_current_user_is_not_readonly
from ..utils import FlaskResponseT

from ..logic.components import get_component_or_none


@frontend.route('/location_types/')
@flask_login.login_required
def location_types() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(403)

    location_types = logic.locations.get_location_types()
    return flask.render_template(
        'location_types/location_types.html',
        location_types=location_types
    )


@frontend.route('/location_types/<int(signed=True):type_id>', methods=['GET', 'POST'])
@flask_login.login_required
def location_type(type_id: int) -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(403)

    try:
        location_type = logic.locations.get_location_type(
            location_type_id=type_id
        )
    except logic.errors.LocationTypeDoesNotExistError:
        return flask.abort(404)

    if flask.request.args.get('mode') == 'edit':
        if location_type.fed_id is not None:
            flask.flash(_('Editing imported location types is not yet supported.'), 'error')
            return flask.abort(403)
        else:
            return show_location_type_form(type_id)

    return flask.render_template(
        'location_types/location_type.html',
        location_type=location_type,
        component=get_component_or_none(location_type.component_id)
    )


@frontend.route('/location_types/new', methods=['GET', 'POST'])
@flask_login.login_required
def new_location_type() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(403)
    return show_location_type_form(None)


class LocationTypeForm(FlaskForm):
    admin_only = BooleanField()
    enable_parent_location = BooleanField()
    enable_sub_locations = BooleanField()
    enable_object_assignments = BooleanField()
    enable_responsible_users = BooleanField()
    enable_instruments = BooleanField()
    show_location_log = BooleanField()
    enable_capacities = BooleanField()


def show_location_type_form(
        type_id: typing.Optional[int]
) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    location_type_form = LocationTypeForm()

    has_error = False
    languages = logic.languages.get_languages()
    location_language_ids = set()
    language_id_by_lang_code = {
        language.lang_code: language.id
        for language in languages
    }
    english_id = logic.languages.Language.ENGLISH
    translation_language_ids = {english_id}
    translated_texts = {
        text_name: {
            'en': ''
        }
        for text_name in ('name', 'location_name_singular', 'location_name_plural')
    }

    if not location_type_form.is_submitted():
        if type_id is not None:
            try:
                location_type = logic.locations.get_location_type(type_id)
            except logic.errors.LocationTypeDoesNotExistError:
                return flask.abort(404)
            # set boolean fields from existing location type
            location_type_form.admin_only.data = location_type.admin_only
            location_type_form.enable_parent_location.data = location_type.enable_parent_location
            location_type_form.enable_sub_locations.data = location_type.enable_sub_locations
            location_type_form.enable_object_assignments.data = location_type.enable_object_assignments
            location_type_form.enable_responsible_users.data = location_type.enable_responsible_users
            location_type_form.enable_instruments.data = location_type.enable_instruments
            location_type_form.enable_capacities.data = location_type.enable_capacities
            location_type_form.show_location_log.data = location_type.show_location_log
            # set translated texts from existing location type
            for text_name in translated_texts:
                existing_translations = getattr(location_type, text_name)
                if existing_translations is not None:
                    for lang_code, translated_text in existing_translations.items():
                        translated_texts[text_name][lang_code] = translated_text
                        if lang_code in language_id_by_lang_code:
                            translation_language_ids.add(language_id_by_lang_code[lang_code])
        else:
            # set boolean field default values
            location_type_form.admin_only.data = False
            location_type_form.enable_parent_location.data = True
            location_type_form.enable_sub_locations.data = True
            location_type_form.enable_object_assignments.data = True
            location_type_form.enable_responsible_users.data = False
            location_type_form.show_location_log.data = False
            location_type_form.enable_instruments.data = True
            location_type_form.enable_capacities.data = False
    else:
        translation_language_id_strs = flask.request.form.getlist('translation-languages')
        translation_language_ids = {
            language_id
            for language_id in language_id_by_lang_code.values()
            if str(language_id) in translation_language_id_strs or language_id == english_id
        }
        # set translated texts from form
        for text_name in translated_texts:
            for language in languages:
                if language.id in translation_language_ids:
                    translated_text = flask.request.form.get(f'{text_name}_{language.id}', '')
                    translated_texts[text_name][language.lang_code] = translated_text
                    location_language_ids.add(language.id)
                    if len(translated_text) == 0 or len(translated_text) > 100:
                        has_error = True

    if not has_error and location_type_form.validate_on_submit():
        if type_id is None:
            type_id = logic.locations.create_location_type(
                admin_only=location_type_form.admin_only.data,
                enable_parent_location=location_type_form.enable_parent_location.data,
                enable_sub_locations=location_type_form.enable_sub_locations.data,
                enable_object_assignments=location_type_form.enable_object_assignments.data,
                enable_responsible_users=location_type_form.enable_responsible_users.data,
                enable_instruments=location_type_form.enable_instruments.data,
                enable_capacities=location_type_form.enable_capacities.data,
                show_location_log=location_type_form.show_location_log.data,
                name=translated_texts['name'],
                location_name_singular=translated_texts['location_name_singular'],
                location_name_plural=translated_texts['location_name_plural']
            ).id
        else:
            logic.locations.update_location_type(
                location_type_id=type_id,
                admin_only=location_type_form.admin_only.data,
                enable_parent_location=location_type_form.enable_parent_location.data,
                enable_sub_locations=location_type_form.enable_sub_locations.data,
                enable_object_assignments=location_type_form.enable_object_assignments.data,
                enable_responsible_users=location_type_form.enable_responsible_users.data,
                enable_instruments=location_type_form.enable_instruments.data,
                enable_capacities=location_type_form.enable_capacities.data,
                show_location_log=location_type_form.show_location_log.data,
                name=translated_texts['name'],
                location_name_singular=translated_texts['location_name_singular'],
                location_name_plural=translated_texts['location_name_plural']
            )
        return flask.redirect(flask.url_for('.location_type', type_id=type_id))

    return flask.render_template(
        'location_types/location_type_form.html',
        location_type_form=location_type_form,
        translation_language_ids=list(translation_language_ids),
        translated_texts=translated_texts,
        ENGLISH=logic.languages.get_language(logic.languages.Language.ENGLISH),
        languages=languages,
        submit_text=_('Create') if type_id is None else _('Save')
    )



# coding: utf-8
"""
Add null if use_real_* is set check to fed_user_aliases.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    constraints = db.session.execute(db.text("""
         SELECT conname
         FROM pg_catalog.pg_constraint
         WHERE conname = 'fed_user_alias_null_if_use_real_data'
    """)).fetchall()

    if len(constraints) > 0:
        return False

    # Perform migration
    db.session.execute(db.text("""
            ALTER TABLE fed_user_aliases
                ADD CONSTRAINT fed_user_alias_null_if_use_real_data
                    CHECK ((name IS NULL OR use_real_name IS FALSE) AND
                        (use_real_email IS FALSE OR email IS NULL) AND
                        (orcid IS NULL OR use_real_orcid IS FALSE) AND
                        (affiliation IS NULL OR use_real_affiliation IS FALSE) AND
                        (role IS NULL OR use_real_role IS FALSE)
                    )
        """))
    return True



# coding: utf-8
"""

"""

import typing

import flask

from . import locations, users
from ..models import Permissions, AllUserLocationPermissions, UserLocationPermissions, GroupLocationPermissions, ProjectLocationPermissions
from ..models.locations import location_user_association_table
from .permissions import ResourcePermissions
from .. import db


__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


location_permissions = ResourcePermissions(
    resource_id_name='location_id',
    all_user_permissions_table=AllUserLocationPermissions,
    anonymous_user_permissions_table=None,
    user_permissions_table=UserLocationPermissions,
    group_permissions_table=GroupLocationPermissions,
    project_permissions_table=ProjectLocationPermissions,
    check_resource_exists=lambda resource_id: locations.check_location_exists(location_id=resource_id)
)


def get_user_location_permissions(
        location_id: int,
        user_id: typing.Optional[int],
        *,
        include_groups: bool = True,
        include_projects: bool = True,
        include_admin_permissions: bool = True
) -> Permissions:
    return get_user_permissions_for_multiple_locations(
        user_id=user_id,
        location_ids=[location_id],
        include_groups=include_groups,
        include_projects=include_projects,
        include_admin_permissions=include_admin_permissions
    )[location_id]


def get_user_permissions_for_multiple_locations(
        location_ids: typing.Sequence[int],
        user_id: typing.Optional[int],
        *,
        include_groups: bool = True,
        include_projects: bool = True,
        include_admin_permissions: bool = True,
        max_permissions: Permissions = Permissions.GRANT
) -> typing.Mapping[int, Permissions]:
    if not location_ids:
        return {}

    additional_permissions = {
        location_id: Permissions.NONE
        for location_id in location_ids
    }
    if user_id is None:
        return {
            location_id: Permissions.NONE
            for location_id in location_ids
        }

    if flask.current_app.config['ONLY_ADMINS_CAN_MANAGE_LOCATIONS']:
        user = users.get_user(user_id)
        # if only admins are allowed to manage locations, only they can have
        # more than READ permissions
        if not (include_admin_permissions and user.has_admin_permissions):
            max_permissions = min(max_permissions, Permissions.READ)

    for location_id in location_ids:
        # apply responsible user permissions
        if bool(db.session.query(db.exists().where(location_user_association_table.c.location_id == location_id, location_user_association_table.c.user_id == user_id)).scalar()):
            additional_permissions[location_id] = min(Permissions.GRANT, max_permissions)

    # resource independent permissions
    return location_permissions.get_permissions_for_user_for_multiple_resources(
        resource_ids=location_ids,
        user_id=user_id,
        include_all_users=True,
        include_groups=include_groups,
        include_projects=include_projects,
        include_admin_permissions=include_admin_permissions,
        limit_readonly_users=True,
        additional_permissions=additional_permissions,
        max_permissions=max_permissions
    )


def get_location_permissions_for_all_users(location_id: int) -> Permissions:
    return location_permissions.get_permissions_for_all_users(resource_id=location_id)


def set_location_permissions_for_all_users(location_id: int, permissions: Permissions) -> None:
    location_permissions.set_permissions_for_all_users(resource_id=location_id, permissions=permissions)


def get_location_permissions_for_users(location_id: int) -> typing.Dict[int, Permissions]:
    return location_permissions.get_permissions_for_users(resource_id=location_id)


def set_user_location_permissions(location_id: int, user_id: int, permissions: Permissions) -> None:
    location_permissions.set_permissions_for_user(resource_id=location_id, user_id=user_id, permissions=permissions)


def get_location_permissions_for_groups(location_id: int) -> typing.Dict[int, Permissions]:
    return location_permissions.get_permissions_for_groups(resource_id=location_id)


def set_group_location_permissions(location_id: int, group_id: int, permissions: Permissions) -> None:
    location_permissions.set_permissions_for_group(resource_id=location_id, group_id=group_id, permissions=permissions)


def get_location_permissions_for_projects(location_id: int) -> typing.Dict[int, Permissions]:
    return location_permissions.get_permissions_for_projects(resource_id=location_id)


def set_project_location_permissions(location_id: int, project_id: int, permissions: Permissions) -> None:
    location_permissions.set_permissions_for_project(resource_id=location_id, project_id=project_id, permissions=permissions)


def get_locations_with_user_permissions(
        user_id: typing.Optional[int],
        permissions: Permissions
) -> typing.List['locations.Location']:
    if user_id is None:
        return []

    all_locations = locations.get_locations()
    if permissions in Permissions.NONE:
        return all_locations
    if permissions not in Permissions.READ:
        user = users.get_user(user_id)
        if user.is_readonly:
            return []
        if not user.is_admin and flask.current_app.config['ONLY_ADMINS_CAN_MANAGE_LOCATIONS']:
            return []
    location_permissions = get_user_permissions_for_multiple_locations(
        location_ids=[location.id for location in all_locations],
        user_id=user_id,
        max_permissions=permissions
    )
    locations_with_user_permissions = []
    for location in all_locations:
        if permissions in location_permissions[location.id]:
            locations_with_user_permissions.append(location)
    return locations_with_user_permissions


def location_is_public(location_id: int) -> bool:
    """
    Returns whether a location is public.

    A location is considered public if its permission settings match those set
    for locations created as public via the frontend, i.e. WRITE permissions
    for all users and no explicitly set permissions for users, groups or
    projects.

    :param location_id: the id of the location
    :return: whether the location is public
    :raise errors.LocationDoesNotExistError: if no location with the given ID
        exists
    """
    if get_location_permissions_for_all_users(location_id) != Permissions.WRITE:
        return False

    if db.session.query(db.exists().where(UserLocationPermissions.location_id == location_id)).scalar():
        return False
    if db.session.query(db.exists().where(GroupLocationPermissions.location_id == location_id)).scalar():
        return False
    if db.session.query(db.exists().where(ProjectLocationPermissions.location_id == location_id)).scalar():
        return False
    locations.check_location_exists(location_id)
    return True



# coding: utf-8
"""

"""
import io
import os
import typing
import zipfile

import flask
import flask_login
import itsdangerous
import werkzeug.utils
from flask_babel import _

from .. import frontend
from ... import logic
from ...logic.objects import get_object
from ...models import Permissions
from ...logic.errors import UserDoesNotExistError, FederationFileNotAvailableError
from .forms import FileForm, FileInformationForm, FileHidingForm, ExternalLinkForm
from ...utils import object_permissions_required, FlaskResponseT
from ..utils import check_current_user_is_not_readonly
from .permissions import on_unauthorized
from ...logic.temporary_files import create_temporary_file, delete_expired_temporary_files


@frontend.route('/objects/<int:object_id>/files/')
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def object_files(object_id: int) -> FlaskResponseT:
    files = logic.files.get_files_for_object(object_id)
    zip_bytes = io.BytesIO()
    with zipfile.ZipFile(zip_bytes, 'w') as zip_file:
        for file in files:
            if file.is_hidden:
                continue
            if file.storage in {'database', 'federation'}:
                try:
                    file_bytes = file.open(read_only=True).read()
                except Exception:
                    pass
                else:
                    zip_file.writestr(os.path.basename(file.original_file_name), file_bytes)
    return flask.Response(
        zip_bytes.getvalue(),
        200,
        headers={
            'Content-Type': 'application/zip',
            'Content-Disposition': f'attachment; filename=object_{object_id}_files.zip'
        }
    )


@frontend.route('/objects/<int:object_id>/files/<int:file_id>', methods=['GET'])
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def object_file(object_id: int, file_id: int) -> FlaskResponseT:
    file = logic.files.get_file_for_object(object_id, file_id)
    if file is None:
        return flask.abort(404)
    if file.is_hidden:
        return flask.abort(403)
    if file.storage in ('database', 'federation'):
        if 'preview' in flask.request.args:
            mime_types = flask.current_app.config.get('MIME_TYPES', {})
            if file.preview_image_binary_data is not None and file.preview_image_mime_type is not None and file.preview_image_mime_type.lower() in mime_types.values():
                return flask.send_file(io.BytesIO(file.preview_image_binary_data), mimetype=file.preview_image_mime_type.lower(), last_modified=file.utc_datetime)
            file_extension = os.path.splitext(file.original_file_name)[1]
            mime_type = mime_types.get(file_extension, None)
            if mime_type is not None:
                try:
                    return flask.send_file(file.open(), mimetype=mime_type, last_modified=file.utc_datetime)
                except FederationFileNotAvailableError:
                    flask.flash(_('File stored in other database is not available.'), 'error')
                    return flask.abort(404)
        try:
            return flask.send_file(file.open(), as_attachment=True, download_name=file.original_file_name, last_modified=file.utc_datetime)
        except FederationFileNotAvailableError:
            flask.flash(_('File stored in other database is not available.'), 'error')
            return flask.abort(404)
    # TODO: better error handling
    return flask.abort(404)


@frontend.route('/objects/<int:object_id>/files/<int:file_id>', methods=['POST'])
@object_permissions_required(Permissions.WRITE)
def update_file_information(object_id: int, file_id: int) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    object = get_object(object_id)
    if object.fed_object_id is not None:
        flask.flash(_('Uploading files for imported objects is not yet supported.'), 'error')
        return flask.abort(403)
    if object.action_id is None:
        return flask.abort(403)
    form = FileInformationForm()
    if form.validate_on_submit():
        title = form.title.data
        url = form.url.data
        description = form.description.data
        try:
            logic.files.update_file_information(
                object_id=object_id,
                file_id=file_id,
                user_id=flask_login.current_user.id,
                title=title,
                url=url,
                description=description
            )
        except logic.errors.FileDoesNotExistError:
            return flask.abort(404)
        return flask.redirect(flask.url_for('.object', object_id=object_id, _anchor=f'file-{file_id}'))
    else:
        if 'url' in form.errors:
            errorcode = form.errors['url']
            return flask.redirect(flask.url_for('.object', object_id=object_id, edit_invalid_link_file=file_id, edit_invalid_link_error=errorcode, _anchor=f'file-{file_id}'))
        return flask.abort(400)


@frontend.route('/objects/<int:object_id>/files/<int:file_id>/hide', methods=['POST'])
@object_permissions_required(Permissions.WRITE)
def hide_file(object_id: int, file_id: int) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    object = get_object(object_id)
    if object.fed_object_id is not None or object.action_id is None:
        return flask.abort(403)
    form = FileHidingForm()
    if not form.validate_on_submit():
        return flask.abort(400)
    reason = form.reason.data
    try:
        logic.files.hide_file(
            object_id=object_id,
            file_id=file_id,
            user_id=flask_login.current_user.id,
            reason=reason
        )
    except logic.errors.FileDoesNotExistError:
        return flask.abort(404)
    flask.flash(_('The file was hidden successfully.'), 'success')
    return flask.redirect(flask.url_for('.object', object_id=object_id, _anchor=f'file-{file_id}'))


@frontend.route('/objects/<int:object_id>/files/mobile_upload/<token>', methods=['GET'])
def mobile_file_upload(object_id: int, token: str) -> FlaskResponseT:
    serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='mobile-upload')
    try:
        user_id, object_id = serializer.loads(token, max_age=15 * 60)
    except itsdangerous.BadSignature:
        return flask.abort(400)
    try:
        user = logic.users.get_user(user_id)
    except UserDoesNotExistError:
        return flask.abort(403)
    if user.is_readonly:
        return flask.abort(403)
    return flask.render_template('mobile_upload.html')


@frontend.route('/objects/<int:object_id>/files/mobile_upload/<token>', methods=['POST'])
def post_mobile_file_upload(object_id: int, token: str) -> FlaskResponseT:
    serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='mobile-upload')
    try:
        user_id, object_id = serializer.loads(token, max_age=15 * 60)
    except itsdangerous.BadSignature:
        return flask.abort(400)
    try:
        user = logic.users.get_user(user_id)
    except UserDoesNotExistError:
        return flask.abort(403)
    if user.is_readonly:
        return flask.abort(403)
    files = flask.request.files.getlist('file_input')
    if not files:
        return flask.redirect(
            flask.url_for(
                '.mobile_file_upload',
                object_id=object_id,
                token=token
            )
        )
    for file_storage in files:
        if file_storage.filename is not None:
            file_name = werkzeug.utils.secure_filename(file_storage.filename)
        else:
            file_name = 'file'
        logic.files.create_database_file(object_id, user_id, file_name, typing.cast(typing.Callable[[typing.BinaryIO], None], lambda stream, file_storage=file_storage: file_storage.save(dst=stream)))
    return flask.render_template('mobile_upload_success.html')


@frontend.route('/objects/<int:object_id>/files/', methods=['POST'])
@object_permissions_required(Permissions.WRITE)
def post_object_files(object_id: int) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    object = get_object(object_id)
    if object.fed_object_id is not None or object.action_id is None:
        return flask.abort(403)
    external_link_form = ExternalLinkForm()
    file_form = FileForm()
    if file_form.validate_on_submit():
        file_source = file_form.file_source.data
        if file_source == 'local':
            files = flask.request.files.getlist(file_form.local_files.name)
            for file_storage in files:
                if file_storage.filename is not None:
                    file_name = werkzeug.utils.secure_filename(file_storage.filename)
                else:
                    file_name = 'file'
                logic.files.create_database_file(object_id, flask_login.current_user.id, file_name, typing.cast(typing.Callable[[typing.BinaryIO], None], lambda stream, file_storage=file_storage: file_storage.save(dst=stream)))
            flask.flash(_('Successfully uploaded files.'), 'success')
        else:
            flask.flash(_('Failed to upload files.'), 'error')
    elif external_link_form.validate_on_submit():
        url = external_link_form.url.data
        logic.files.create_url_file(object_id, flask_login.current_user.id, url)
        flask.flash(_('Successfully posted link.'), 'success')
    elif external_link_form.errors:
        errorcode = 1
        if 'url' in external_link_form.errors:
            errorcode = external_link_form.errors['url']

        flask.flash(_('Failed to post link.'), 'error')
        return flask.redirect(flask.url_for('.object', object_id=object_id, invalid_link_error=errorcode, _anchor='anchor-post-link'))
    else:
        flask.flash(_('Failed to upload files.'), 'error')
    return flask.redirect(flask.url_for('.object', object_id=object_id))


@frontend.route('/objects/temporary_files/', methods=['POST'])
@flask_login.login_required
def temporary_file_upload() -> FlaskResponseT:
    delete_expired_temporary_files()
    context_id_token = flask.request.form.get('context_id_token', default='')
    if not context_id_token:
        return flask.abort(400)
    serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='temporary-file-upload')
    try:
        user_id, context_id = serializer.loads(context_id_token, max_age=flask.current_app.config['TEMPORARY_FILE_TIME_LIMIT'])
    except itsdangerous.BadData:
        return flask.abort(400)
    if user_id != flask_login.current_user.id:
        return flask.abort(400)
    files = flask.request.files.getlist('file')
    if len(files) != 1:
        return flask.abort(400)
    file = files[0]
    file_name = file.filename or ''
    binary_data = file.stream.read()
    temporary_file = create_temporary_file(
        context_id=context_id,
        file_name=file_name,
        user_id=user_id,
        binary_data=binary_data
    )
    return str(temporary_file.id)



# coding: utf-8
"""

"""
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .groups import Group
from .users import User
from .objects import Objects
from .permissions import Permissions
from .utils import Model

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class UserObjectPermissions(Model):
    __tablename__ = 'user_object_permissions'

    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["UserObjectPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(object_id, user_id),
    )


class GroupObjectPermissions(Model):
    __tablename__ = 'group_object_permissions'

    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    group_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["GroupObjectPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(object_id, group_id),
    )


class ProjectObjectPermissions(Model):
    __tablename__ = 'project_object_permissions'

    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    project_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('projects.id', ondelete="CASCADE"), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ProjectObjectPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(object_id, project_id),
    )


class AllUserObjectPermissions(Model):
    __tablename__ = 'all_user_object_permissions'

    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), primary_key=True)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["AllUserObjectPermissions"]]


class AnonymousUserObjectPermissions(Model):
    __tablename__ = 'anonymous_user_object_permissions'

    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), primary_key=True)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["AnonymousUserObjectPermissions"]]



import base64
import copy
import datetime
import dataclasses
import functools
import secrets
import typing
import urllib.parse

import bcrypt
import flask
from fido2.server import Fido2Server
from fido2.webauthn import AttestationConveyancePreference, PublicKeyCredentialRpEntity, AttestedCredentialData

from .. import logic, db
from .ldap import validate_user, create_user_from_ldap, is_ldap_configured
from ..models import Authentication, AuthenticationType, TwoFactorAuthenticationMethod, HTTPMethod, FederatedIdentity
from . import errors, api_log


# number of rounds for generating new salts for hashing passwords with crypt
# 12 is the default in the Python bcrypt module, this variable allows
# overriding this in tests
NUM_BCYRPT_ROUNDS = 12
# maximum password length (in bytes) supported by bcrypt.hashpw
# versions <5.0.0 have truncated passwords to that length before, but starting
# with bcrypt 5.0.0 a ValueError is raised instead
MAX_BCRYPT_PASSWORD_LENGTH = 72


@dataclasses.dataclass(frozen=True)
class AuthenticationMethod:
    """
    This class provides an immutable wrapper around models.authentication.Authentication.
    """
    id: int
    user_id: int
    login: typing.Dict[str, typing.Any]
    type: AuthenticationType
    confirmed: bool
    user: logic.users.User

    @classmethod
    def from_database(cls, authentication_method: Authentication) -> 'AuthenticationMethod':
        return AuthenticationMethod(
            id=authentication_method.id,
            user_id=authentication_method.user_id,
            login=copy.deepcopy(authentication_method.login),
            type=authentication_method.type,
            confirmed=authentication_method.confirmed,
            user=logic.users.get_user(authentication_method.user_id)
        )


def _hash_password(password: str) -> str:
    # manually truncate long passwords to ensure consistent behavior to bcrypt versions <5.0.0
    return str(bcrypt.hashpw(password.encode('utf-8')[:MAX_BCRYPT_PASSWORD_LENGTH], bcrypt.gensalt(rounds=NUM_BCYRPT_ROUNDS)).decode('utf-8'))


def _validate_password_hash(password: str, password_hash: str) -> bool:
    # manually truncate long passwords to ensure consistent behavior to bcrypt versions <5.0.0
    return bool(bcrypt.checkpw(password.encode('utf-8')[:MAX_BCRYPT_PASSWORD_LENGTH], password_hash.encode('utf-8')))


def _validate_password_authentication(authentication_method: Authentication, password: str) -> bool:
    password_hash = authentication_method.login['bcrypt_hash']
    return _validate_password_hash(password, password_hash)


def _add_password_authentication(user_id: int, login: str, password: str, authentication_type: AuthenticationType, confirmed: bool = True) -> None:
    login = login.lower().strip()
    if check_authentication_method_with_login_exists(login=login):
        raise errors.AuthenticationMethodAlreadyExists('An authentication method with this login already exists')
    authentication = Authentication(
        login={'login': login, 'bcrypt_hash': _hash_password(password)},
        authentication_type=authentication_type,
        confirmed=confirmed,
        user_id=user_id
    )
    db.session.add(authentication)
    db.session.commit()


def add_other_authentication(user_id: int, name: str, password: str, confirmed: bool = True) -> None:
    """
    Add an authentication method of type OTHER for a given user.

    :param user_id: the ID of an existing user
    :param name: the name to use during authentication
    :param password: the password
    :param confirmed: whether the authentication method has been confirmed already
    """
    _add_password_authentication(user_id, name, password, AuthenticationType.OTHER, confirmed)


def add_email_authentication(user_id: int, email: str, password: str, confirmed: bool = True) -> None:
    """
    Add an authentication method of type EMAIL for a given user.

    :param user_id: the ID of an existing user
    :param email: the email to use during authentication
    :param password: the password
    :param confirmed: whether the authentication method has been confirmed already
    :raise errors.AuthenticationMethodWrong: when email is no valid email
        address
    """
    email = email.lower().strip()
    if '@' not in email[1:-1]:
        raise errors.AuthenticationMethodWrong('Login must be a valid email address')
    _add_password_authentication(user_id, email, password, AuthenticationType.EMAIL, confirmed)
    if not confirmed:
        logic.utils.send_email_confirmation_email(
            email=email,
            user_id=user_id,
            salt='add_login'
        )


def add_ldap_authentication(user_id: int, ldap_uid: str, password: str, confirmed: bool = True) -> None:
    """
    Add an authentication method of type LDAP for a given user.

    :param user_id: the ID of an existing user
    :param ldap_uid: the LDAP uid to use during authentication
    :param password: the LDAP password
    :param confirmed: whether the authentication method has been confirmed already
    :raise errors.AuthenticationMethodAlreadyExists: when an LDAP authentication
        method with the given UID already exists
    :raise errors.LDAPAccountAlreadyExistError: when an LDAP authentication
        method already exists for this user
    :raise errors.LDAPAccountOrPasswordWrongError: when the UID and password
        combination could not be validated
    """
    ldap_uid = ldap_uid.lower().strip()
    if Authentication.query.filter(Authentication.login['login'].astext == ldap_uid).first():
        raise errors.AuthenticationMethodAlreadyExists('An authentication method with this login already exists')
    if Authentication.query.filter(Authentication.type == AuthenticationType.LDAP, Authentication.user_id == user_id).first():
        raise errors.LDAPAccountAlreadyExistError('An LDAP-based authentication method already exists for this user')
    if not validate_user(ldap_uid, password):
        raise errors.LDAPAccountOrPasswordWrongError('Ldap login or password wrong')
    authentication = Authentication(
        login={'login': ldap_uid},
        authentication_type=AuthenticationType.LDAP,
        confirmed=confirmed,
        user_id=user_id
    )
    db.session.add(authentication)
    db.session.commit()


def add_api_token(user_id: int, api_token: str, description: str) -> None:
    """
    Add an API token as an authentication method for a given user.

    :param user_id: the ID of an existing user
    :param api_token: the api token as a 64 character string
    :param description: a description for the token
    """

    assert len(api_token) == 64
    api_token = api_token.lower()
    # split into a short part (8 hex digits / 4 bytes) for identification and a long part for authentication
    login, password = api_token[:8], api_token[8:]
    authentication = Authentication(
        login={
            'login': login,
            'bcrypt_hash': _hash_password(password),
            'description': description
        },
        authentication_type=AuthenticationType.API_TOKEN,
        confirmed=True,
        user_id=user_id
    )
    db.session.add(authentication)
    db.session.commit()


ALL_AUTHENTICATION_TYPES = {
    authentication_type
    for authentication_type in AuthenticationType.__members__.values()
}


def get_authentication_methods(
        user_id: int,
        authentication_types: typing.Set[AuthenticationType]
) -> typing.Sequence[AuthenticationMethod]:
    """
    Get all authentication methods for a given user.

    :param user_id: the ID of an existing user
    :param authentication_types: the authentication types to filter for
    :return: the authentication methods for the user
    """
    authentication_methods = Authentication.query.filter(
        Authentication.user_id == user_id,
        Authentication.type.in_(authentication_types)
    ).all()
    return [
        AuthenticationMethod.from_database(authentication_method)
        for authentication_method in authentication_methods
    ]


def get_api_tokens(user_id: int) -> typing.Sequence[AuthenticationMethod]:
    """
    Get all API tokens for a given user.

    :param user_id: the ID of an existing user
    :return: the API tokens for the user
    """
    return get_authentication_methods(
        user_id=user_id,
        authentication_types={AuthenticationType.API_TOKEN}
    )


def get_authentication_method(authentication_method_id: int) -> AuthenticationMethod:
    """
    Get an authentication method by its ID.

    :param authentication_method_id: the ID of an existing authentication method
    :return: the authentication method
    :raise errors.AuthenticationMethodDoesNotExistError: if the authentication method does not exist
    """
    authentication_method = Authentication.query.filter_by(id=authentication_method_id).first()
    if authentication_method is None:
        raise errors.AuthenticationMethodDoesNotExistError()
    return AuthenticationMethod.from_database(authentication_method)


def check_authentication_method_with_login_exists(login: str) -> bool:
    """
    Return whether an authentication method with a given login exists.

    :param login: the login to check
    :return: whether an authentication method with the login exists
    """
    return bool(db.session.query(db.exists().where(Authentication.login['login'].astext == login)).scalar())


def confirm_authentication_method_by_email(
        user_id: int,
        email: str
) -> None:
    """
    Mark an authentication method as confirmed by its associated email.

    :param user_id: the ID of an existing user
    :param email: the email for this authentication method
    :raise errors.AuthenticationMethodDoesNotExistError: if the authentication method does not exist
    """
    authentication_method = Authentication.query.filter(
        Authentication.user_id == user_id,
        Authentication.login['login'].astext == email
    ).first()
    if authentication_method is None:
        raise errors.AuthenticationMethodDoesNotExistError()
    authentication_method.confirmed = True
    db.session.add(authentication_method)
    db.session.commit()


def add_fido2_passkey(user_id: int, credential_data: AttestedCredentialData, description: str) -> None:
    """
    Add a FIDO2 passkey as an authentication method for a given user.

    :param user_id: the ID of an existing user
    :param credential_data: the passkey credential data
    :param description: a description for the passkey
    """
    authentication = Authentication(
        login={
            'credential_id': base64.b64encode(credential_data.credential_id).decode('utf-8'),
            'credential_data': base64.b64encode(credential_data).decode('utf-8'),
            'description': description
        },
        authentication_type=AuthenticationType.FIDO2_PASSKEY,
        confirmed=True,
        user_id=user_id
    )
    db.session.add(authentication)
    db.session.commit()


def get_all_fido2_passkey_credentials() -> typing.Sequence[AttestedCredentialData]:
    """
    Get all known FIDO2 passkey credentials.

    :return: a list of credentials
    """
    return [
        AttestedCredentialData(base64.b64decode(authentication.login['credential_data'].encode('utf-8')))
        for authentication in Authentication.query.filter_by(type=AuthenticationType.FIDO2_PASSKEY).all()
    ]


def get_user_id_for_fido2_passkey_credential_id(credential_id: bytes) -> typing.Optional[int]:
    """
    Get the user ID for a given FIDO2 passkey credential ID.

    :return: the user ID, or None if the credential ID is unknown
    """
    authentication = Authentication.query.filter(
        db.and_(
            Authentication.type == AuthenticationType.FIDO2_PASSKEY,
            Authentication.login['credential_id'].astext == base64.b64encode(credential_id).decode('utf-8')
        )
    ).first()
    if authentication is None:
        return None
    return authentication.user_id


def add_federated_login_authentication(federated_identity: FederatedIdentity) -> None:
    """
    Add a federated login authentication method for a given federated identity.

    :param federated_identity: the federated identity
    """
    if Authentication.query.filter(
        Authentication.type == AuthenticationType.FEDERATED_LOGIN,
        Authentication.user_id == federated_identity.user_id,
        Authentication.login['fed_user_id'].astext.cast(db.Integer) == federated_identity.local_fed_user.fed_id,
        Authentication.login['component_id'].astext.cast(db.Integer) == federated_identity.local_fed_user.component_id
    ).first():
        raise errors.AuthenticationMethodAlreadyExists('An authentication method for this federated identity already exists')

    authentication = Authentication(
        login={
            'fed_user_id': federated_identity.local_fed_user.fed_id,
            'component_id': federated_identity.local_fed_user.component_id
        },
        authentication_type=AuthenticationType.FEDERATED_LOGIN,
        user_id=federated_identity.user_id,
        confirmed=True
    )

    db.session.add(authentication)
    db.session.commit()


def add_oidc_authentication(user_id: int, iss: str, sub: str) -> None:
    """
    Add an authentication method of type OIDC for a given user.

    :param user_id: the ID of an existing user
    :param iss: the issuer
    :param sub: the subject identifier
    """
    authentication = Authentication(
        login={
            'iss': iss,
            'sub': sub,
        },
        authentication_type=AuthenticationType.OIDC,
        confirmed=True,
        user_id=user_id
    )
    db.session.add(authentication)
    db.session.commit()


def get_oidc_authentications_by_sub(iss: str, sub: str) -> typing.List[AuthenticationMethod]:
    """
    Get all authentication methods of type OIDC with the given issuer and
    subject identifier.

    :param iss: the issuer
    :param sub: the subject identifier
    """
    return [
        AuthenticationMethod.from_database(method)
        for method in
        Authentication.query.filter(
            db.and_(
                Authentication.type == AuthenticationType.OIDC,
                Authentication.login['iss'].astext == iss,
                Authentication.login['sub'].astext == sub,
            )
        ).all()
    ]


@functools.cache
def _verify_origin(origin: str) -> bool:
    origin_parsed = urllib.parse.urlparse(origin)
    index_parsed = urllib.parse.urlparse(flask.url_for('.index', _external=True))
    return origin_parsed.scheme == index_parsed.scheme and origin_parsed.netloc == index_parsed.netloc


def get_webauthn_server() -> Fido2Server:
    """
    Get the FIDO2 webauthn server.

    :return: a Fido2Server instance
    """
    rp = PublicKeyCredentialRpEntity(
        name=flask.current_app.config['SERVICE_NAME'],
        id=(flask.current_app.config['SERVER_NAME'] or '').split(':')[0]
    )
    return Fido2Server(
        rp,
        attestation=AttestationConveyancePreference.NONE,
        verify_origin=_verify_origin
    )


def login(login: str, password: str) -> typing.Optional[logic.users.User]:
    """
    Authenticate a user and create an LDAP based user if necessary.

    :param login: the name, email or LDAP uid to use during authentication
    :param password: the password
    :return: the user or None
    """
    # convert to lower case to enforce case insensitivity
    login = login.lower().strip()
    # filter email + password or username + password or username (ldap)
    authentication_methods = Authentication.query.filter(
        db.or_(
            db.and_(Authentication.login['login'].astext == login,
                    Authentication.type == AuthenticationType.EMAIL),
            db.and_(Authentication.login['login'].astext == login,
                    Authentication.type == AuthenticationType.LDAP),
            db.and_(Authentication.login['login'].astext == login,
                    Authentication.type == AuthenticationType.OTHER)
        )
    ).all()

    for authentication_method in authentication_methods:
        if not authentication_method.confirmed:
            continue
        if authentication_method.type == AuthenticationType.LDAP and is_ldap_configured():
            if validate_user(login, password):
                return logic.users.User.from_database(authentication_method.user)
        elif authentication_method.type in {AuthenticationType.EMAIL, AuthenticationType.OTHER}:
            if _validate_password_authentication(authentication_method, password):
                return logic.users.User.from_database(authentication_method.user)

    # no matching authentication method in db
    if not authentication_methods and '@' not in login and is_ldap_configured():
        # try to authenticate against ldap, if login is no email
        if not validate_user(login, password):
            return None

        user = create_user_from_ldap(login)
        if user is None:
            return None
        try:
            add_ldap_authentication(user.id, login, password)
        except errors.AuthenticationMethodAlreadyExists:
            # user might have been created in the background already due to concurrent, duplicate requests
            pass
        return user
    return None


def login_via_api_token(api_token: str) -> typing.Optional[logic.users.User]:
    """
    Authenticate a user using an API token.

    :param api_token: the API token to use during authentication
    :return: the user or None
    """
    # convert to lower case to enforce case insensitivity
    api_token = api_token.lower().strip()
    username, password = api_token[:8], api_token[8:]
    authentication_methods = Authentication.query.filter(
        db.and_(Authentication.login['login'].astext == username,
                Authentication.type == AuthenticationType.API_TOKEN)
    ).all()

    for authentication_method in authentication_methods:
        if not authentication_method.confirmed:
            continue
        if _validate_password_authentication(authentication_method, password):
            api_log.create_log_entry(authentication_method.id, HTTPMethod.from_name(flask.request.method), flask.request.path)
            return logic.users.User.from_database(authentication_method.user)
    return None


def generate_api_access_token(
        user_id: int,
        description: typing.Optional[str]
) -> typing.Dict[str, typing.Optional[str]]:
    """
    Create a new API access token as an authentication method for a given user.

    :param user_id: the ID of an existing user
    :param description: an optional description
    :return: a dict containing the access token information
    """
    expiration_utc_datetime = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=1)
    expiration_utc_datetime_str = expiration_utc_datetime.strftime('%Y-%m-%d %H:%M:%S')

    access_token = secrets.token_hex(32)
    refresh_token = secrets.token_hex(32)

    refresh_token_login, refresh_token_password = refresh_token[:8], refresh_token[8:]
    db.session.add(Authentication(
        login={
            'access_token': access_token,
            'refresh_token_login': refresh_token_login,
            'refresh_token_hash': _hash_password(refresh_token_password),
            'expiration_utc_datetime': expiration_utc_datetime_str,
            'description': description
        },
        authentication_type=AuthenticationType.API_ACCESS_TOKEN,
        confirmed=True,
        user_id=user_id
    ))
    db.session.commit()
    return {
        'access_token': access_token,
        'refresh_token': refresh_token,
        'expiration_utc_datetime': expiration_utc_datetime_str,
        'description': description
    }


def remove_expired_api_access_tokens() -> None:
    """
    Delete all expired API access tokens.
    """
    authentication_methods = Authentication.query.filter_by(
        type=AuthenticationType.API_ACCESS_TOKEN
    ).all()
    current_utc_datetime = datetime.datetime.now(datetime.timezone.utc)
    for authentication_method in authentication_methods:
        expiration_utc_datetime_str = authentication_method.login['expiration_utc_datetime']
        expiration_utc_datetime = datetime.datetime.strptime(
            expiration_utc_datetime_str,
            '%Y-%m-%d %H:%M:%S'
        ).replace(tzinfo=datetime.timezone.utc)
        if expiration_utc_datetime <= current_utc_datetime:
            db.session.delete(authentication_method)
    db.session.commit()


def refresh_api_access_token(api_refresh_token: str) -> typing.Optional[typing.Dict[str, str]]:
    """
    Replace an API access token using its refresh token.

    :param api_refresh_token: the refresh token
    :return: a dict containing the new access token information
    """
    remove_expired_api_access_tokens()
    api_refresh_token = api_refresh_token.lower().strip()
    refresh_token_login, refresh_token_password = api_refresh_token[:8], api_refresh_token[8:]
    authentication_methods = Authentication.query.filter(
        db.and_(
            Authentication.login['refresh_token_login'].astext == refresh_token_login,
            Authentication.type == AuthenticationType.API_ACCESS_TOKEN
        )
    ).all()
    expiration_utc_datetime = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=1)
    expiration_utc_datetime_str = expiration_utc_datetime.strftime('%Y-%m-%d %H:%M:%S')

    for authentication_method in authentication_methods:
        if not authentication_method.confirmed:
            continue
        if _validate_password_hash(refresh_token_password, authentication_method.login['refresh_token_hash']):
            new_access_token = secrets.token_hex(32)
            new_refresh_token = secrets.token_hex(32)
            refresh_token_login, refresh_token_password = new_refresh_token[:8], new_refresh_token[8:]
            description = authentication_method.login['description']
            authentication_method.login = {
                'access_token': new_access_token,
                'refresh_token_login': refresh_token_login,
                'refresh_token_hash': _hash_password(refresh_token_password),
                'expiration_utc_datetime': expiration_utc_datetime_str,
                'description': description
            }
            db.session.add(authentication_method)
            db.session.commit()
            return {
                'access_token': new_access_token,
                'refresh_token': new_refresh_token,
                'expiration_utc_datetime': expiration_utc_datetime_str,
                'description': description
            }
    return None


def login_via_api_access_token(api_access_token: str) -> typing.Optional[logic.users.User]:
    """
    Authenticate a user using an API access token.

    :param api_access_token: the API access token to use for authentication
    :return: the user, or None
    """
    # convert to lower case to enforce case insensitivity
    api_access_token = api_access_token.lower().strip()
    remove_expired_api_access_tokens()
    authentication_methods = Authentication.query.filter(
        db.and_(
            Authentication.login['access_token'].astext == api_access_token,
            Authentication.type == AuthenticationType.API_ACCESS_TOKEN
        )
    ).all()

    for authentication_method in authentication_methods:
        if not authentication_method.confirmed:
            continue
        api_log.create_log_entry(authentication_method.id, HTTPMethod.from_name(flask.request.method), flask.request.path)
        return logic.users.User.from_database(authentication_method.user)
    return None


def login_via_api_refresh_token(api_refresh_token: str) -> typing.Optional[logic.users.User]:
    """
    Authenticate a user using a refresh token.

    :param api_refresh_token: the refresh token to use for authentication
    :return: the user, or None
    """
    # convert to lower case to enforce case insensitivity
    api_refresh_token = api_refresh_token.lower().strip()
    refresh_token_login, refresh_token_password = api_refresh_token[:8], api_refresh_token[8:]
    remove_expired_api_access_tokens()
    authentication_methods = Authentication.query.filter(
        db.and_(
            Authentication.login['refresh_token_login'].astext == refresh_token_login,
            Authentication.type == AuthenticationType.API_ACCESS_TOKEN
        )
    ).all()

    for authentication_method in authentication_methods:
        if not authentication_method.confirmed:
            continue
        if _validate_password_hash(refresh_token_password, authentication_method.login['refresh_token_hash']):
            api_log.create_log_entry(authentication_method.id, HTTPMethod.from_name(flask.request.method), flask.request.path)
            return logic.users.User.from_database(authentication_method.user)
    return None


def login_via_oidc_access_token(access_token: str) -> typing.Optional[logic.users.User]:
    """
    Authenticate a user using an OIDC Access Token, if OIDC is configured and
    Access Tokens may be used as API keys.

    :param access_token: the access token to use for authentication
    :return: the user, or None
    """
    if not flask.current_app.config['OIDC_ACCESS_TOKEN_AS_API_KEY']:
        return None
    try:
        authentication_method, user = logic.oidc.validate_access_token(access_token)
    except Exception:
        return None
    api_log.create_log_entry(authentication_method.id, HTTPMethod.from_name(flask.request.method), flask.request.path)
    return user


def add_authentication_method(user_id: int, login: str, password: str, authentication_type: AuthenticationType) -> bool:
    """
    Add an authentication method for a user.

    :param user_id: the ID of an existing user
    :param login: the name, email or LDAP uid to use
    :param password: the password
    :param authentication_type: the type of authentication
    :return: whether the authentication method was added
    :raise errors.AuthenticationMethodAlreadyExists: if the authentication method exists already
    :raise errors.AuthenticationMethodWrong: if the authentication method was used in a wrong way
    :raise errors.LdapAccountAlreadyExist: if a user with this LDAP account already exists
    :raise errors.LdapAccountOrPasswordWrong: if the LDAP account and password could not be validated
    """
    if password is None or password == '':
        return False
    if login is None or login == '':
        return False
    # convert to lower case to enforce case insensitivity
    login = login.lower().strip()

    if authentication_type == AuthenticationType.LDAP:
        add_ldap_authentication(user_id, login, password)
        return True
    if authentication_type == AuthenticationType.EMAIL:
        add_email_authentication(user_id, login, password, False)
        return True
    return False


def remove_authentication_method(authentication_method_id: int) -> bool:
    """
    Remove an authentication method.

    :param authentication_method_id: the ID of an existing authentication method
    :raise errors.OnlyOneAuthenticationMethod: when this would remove the only
        authentication method for this user
    """
    authentication_method = Authentication.query.filter(Authentication.id == authentication_method_id).first()
    if authentication_method is None:
        return False
    if authentication_method.type not in {AuthenticationType.API_TOKEN, AuthenticationType.API_ACCESS_TOKEN}:
        authentication_methods_query = Authentication.query.filter(Authentication.user_id == authentication_method.user_id, Authentication.type != AuthenticationType.API_TOKEN, Authentication.type != AuthenticationType.API_ACCESS_TOKEN)
        if not flask.current_app.config['ENABLE_FEDERATED_LOGIN']:
            authentication_methods_query = authentication_methods_query.filter(Authentication.type != AuthenticationType.FEDERATED_LOGIN)
        authentication_methods_count = authentication_methods_query.count()
        if authentication_methods_count <= 1:
            raise errors.OnlyOneAuthenticationMethod('one authentication-method must at least exist, delete not possible')
    db.session.delete(authentication_method)
    db.session.commit()
    return True


def change_password_in_authentication_method(authentication_method_id: int, password: str) -> bool:
    """
    Change the password for a password-based authentication method.

    :param authentication_method_id: the ID of an authentication method
    :param password: the new password
    :return: whether the password was changed
    """
    if password is None or password == '':
        return False
    if authentication_method_id is None or authentication_method_id <= 0:
        return False
    authentication_method = Authentication.query.filter(Authentication.id == authentication_method_id).first()
    if authentication_method is None:
        return False
    if authentication_method.type not in {AuthenticationType.EMAIL, AuthenticationType.OTHER}:
        return False
    authentication_method.login = {'login': authentication_method.login['login'], 'bcrypt_hash': _hash_password(password)}
    db.session.add(authentication_method)
    db.session.commit()
    return True


def is_login_available(login: str) -> bool:
    """
    Return whether or not a login is still available for use.

    For password-based authentication, the login can only be used once.

    :param login: the login for which to check availability
    :return: whether the login is available
    """
    return Authentication.query.filter(Authentication.login['login'].astext == login).first() is None


def _create_two_factor_authentication_method(
        user_id: int,
        data: typing.Dict[str, typing.Any],
        active: bool = False
) -> TwoFactorAuthenticationMethod:
    """
    Create a new two-factor authentication method.

    :param user_id: the ID of an existing user
    :param data: the method data
    :param active: whether the new method should be active
    :return: the newly created method
    """
    two_factor_authentication_method = TwoFactorAuthenticationMethod(
        user_id=user_id,
        data=data,
        active=active
    )
    db.session.add(two_factor_authentication_method)
    db.session.commit()
    return two_factor_authentication_method


def get_two_factor_authentication_methods(
        user_id: int,
        active: typing.Optional[bool] = None
) -> typing.List[TwoFactorAuthenticationMethod]:
    """
    Get all two-factor authentication methods for a user.

    :param user_id: the ID of an existing user
    :param active: whether the queries need to be active, or None
    :return: a list containing all methods
    """
    query = TwoFactorAuthenticationMethod.query.filter_by(user_id=user_id)
    if active is not None:
        query = query.filter_by(active=active)
    return query.order_by(TwoFactorAuthenticationMethod.id).all()


def get_active_two_factor_authentication_methods(
        user_id: int
) -> typing.Sequence[TwoFactorAuthenticationMethod]:
    """
    Get the currently active two-factor authentication methods.

    :param user_id: the ID of an existing user
    :return: a list of active methods
    """
    return get_two_factor_authentication_methods(user_id=user_id, active=True)


def _set_two_factor_authentication_method_active(
        id: int,
        active: bool
) -> None:
    method = TwoFactorAuthenticationMethod.query.filter_by(id=id).first()
    if method is None:
        raise errors.TwoFactorAuthenticationMethodDoesNotExistError()
    method.active = active
    db.session.add(method)
    db.session.commit()


def activate_two_factor_authentication_method(
        id: int
) -> None:
    """
    Activate a two-factor authentication method.

    :param id: the ID of the method
    :raise errors.TwoFactorAuthenticationMethodDoesNotExistError: if no such method exists
    """
    _set_two_factor_authentication_method_active(id, active=True)


def deactivate_two_factor_authentication_method(
        id: int
) -> None:
    """
    Deactivate a two-factor authentication method.

    :param id: the ID of the method
    :raise errors.TwoFactorAuthenticationMethodDoesNotExistError: if no such method exists
    """
    _set_two_factor_authentication_method_active(id, active=False)


def delete_two_factor_authentication_method(
        id: int
) -> None:
    """
    Delete a two-factor authentication method.

    :param id: the ID of the method
    :raise errors.TwoFactorAuthenticationMethodDoesNotExistError: if no such method exists
    """
    deleted_method = TwoFactorAuthenticationMethod.query.filter_by(id=id).first()
    if deleted_method is None:
        raise errors.TwoFactorAuthenticationMethodDoesNotExistError()
    db.session.delete(deleted_method)
    db.session.commit()


def create_totp_two_factor_authentication_method(
        user_id: int,
        secret: str,
        description: str = ''
) -> TwoFactorAuthenticationMethod:
    """
    Create a new TOTP-based two-factor authentication method.

    :param user_id: the ID of an existing user
    :param secret: the TOTP secret
    :param description: a description
    :return: the newly created method
    """
    return _create_two_factor_authentication_method(
        user_id=user_id,
        data={
            'type': 'totp',
            'secret': secret,
            'description': description.strip()
        }
    )


def get_all_two_factor_fido2_passkey_credentials_for_user(
        user_id: int
) -> typing.Sequence[AttestedCredentialData]:
    """
    Get all FIDO2 passkey credentials for two-factor authentication methods for a given user.

    :param user_id: the ID of an existing user
    :return: a list of credentials
    """
    return [
        AttestedCredentialData(base64.b64decode(authentication_method.data['credential_data'].encode('utf-8')))
        for authentication_method in get_two_factor_authentication_methods(user_id)
        if authentication_method.data.get('type') == 'fido2_passkey'
    ]


def create_fido2_passkey_two_factor_authentication_method(
        user_id: int,
        credential_data: AttestedCredentialData,
        description: str = ''
) -> TwoFactorAuthenticationMethod:
    """
    Create a new FIDO2 Passkey two-factor authentication method.

    :param user_id: the ID of an existing user
    :param credential_data: the credential data
    :param description: a description
    :return: the newly created method
    """
    return _create_two_factor_authentication_method(
        user_id=user_id,
        data={
            'type': 'fido2_passkey',
            'credential_data': base64.b64encode(credential_data).decode('utf-8'),
            'description': description.strip()
        }
    )



import typing

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import Mapped, Query, relationship

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .groups import Group
    from .projects import Project

basic_group_category_association_table = db.Table(
    "basic_group_category_association_table",
    db.metadata,
    db.Column("group_category_id", db.ForeignKey("group_categories.id", ondelete="CASCADE")),
    db.Column("basic_group_id", db.ForeignKey("groups.id", ondelete="CASCADE")),
)

project_group_category_association_table = db.Table(
    "project_group_category_association_table",
    db.metadata,
    db.Column("group_category_id", db.ForeignKey("group_categories.id", ondelete="CASCADE")),
    db.Column("project_group_id", db.ForeignKey("projects.id", ondelete="CASCADE")),
)


class GroupCategory(Model):
    __tablename__ = 'group_categories'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    name: Mapped[typing.Dict[str, str]] = db.Column(postgresql.JSON, nullable=False)
    parent_category_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('group_categories.id', ondelete="CASCADE"), nullable=True)
    basic_groups: Mapped[typing.List['Group']] = relationship("Group", secondary=basic_group_category_association_table, back_populates="categories")
    project_groups: Mapped[typing.List['Project']] = relationship("Project", secondary=project_group_category_association_table, back_populates="categories")

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["GroupCategory"]]



# coding: utf-8
"""
Permissions

Various resources in SampleDB, such as objects and actions, have a unified
permission system that controls how each user can interact with them.

There are three general types of permissions for each resource:
 - READ, being able to access the resource and its information,
 - WRITE, being able to make changes to the resources, and
 - GRANT, being able to grant permissions to other users.

A user may get permissions in several ways:
- they may be implicitly granted permissions along with all other users
- they may be explicitly granted permissions as individual users
- they may be transitively granted permissions as member of a group
- they may be transitively granted permissions as member of a project
- they may be an administrator utilizing administrator permissions

Depending on the type of resource, there may be additional ways of getting
permissions, e.g. by being an instrument responsible user.

Even if a user has been granted WRITE or GRANT permissions in some of the ways
listed above, they might be limited to READ permissions only if they have been
marked as read only users.

The ResourcePermissions class in this module is meant to be used by modules
for the individual resource types, as a base to avoid code duplication.
"""
import typing

from . import users, groups, projects
from ..models import Permissions
from .. import db


class ResourcePermissions:
    def __init__(
            self,
            resource_id_name: str,
            all_user_permissions_table: typing.Any,
            anonymous_user_permissions_table: typing.Any,
            user_permissions_table: typing.Any,
            group_permissions_table: typing.Any,
            project_permissions_table: typing.Any,
            check_resource_exists: typing.Callable[[int], typing.Any]
    ) -> None:
        self._resource_id_name = resource_id_name
        self._all_user_permissions_table = all_user_permissions_table
        self._anonymous_user_permissions_table = anonymous_user_permissions_table
        self._user_permissions_table = user_permissions_table
        self._group_permissions_table = group_permissions_table
        self._project_permissions_table = project_permissions_table
        self._check_resource_exists = check_resource_exists

    def _get_user_independent_permissions(
            self,
            resource_id: int,
            table: typing.Any
    ) -> Permissions:
        """
        Return the permissions for a specific resource from a user independent table.

        :param resource_id: the ID of an existing resource
        :param table: the permissions table
        :return: the user independent permissions
        """
        if table is None:
            return Permissions.NONE
        permissions = table.query.filter_by(**{self._resource_id_name: resource_id}).first()
        if permissions is None:
            self._check_resource_exists(resource_id)
            return Permissions.NONE
        return typing.cast(Permissions, permissions.permissions)

    def _get_user_independent_permissions_for_multiple_resources(
            self,
            resource_ids: typing.Sequence[int],
            table: typing.Any
    ) -> typing.Mapping[int, Permissions]:
        """
        Return the permissions for a list of resources from a user independent table.

        :param resource_ids: the list of IDs of existing resources
        :param table: the permissions table
        :return: the user independent permissions
        """
        if table is None:
            return {
                resource_id: Permissions.NONE
                for resource_id in resource_ids
            }
        independent_permissions = {}
        resource_id_column = getattr(table, self._resource_id_name)
        permissions_rows = table.query.filter(resource_id_column.in_(resource_ids)).all()
        for permissions_row in permissions_rows:
            resource_id = getattr(permissions_row, self._resource_id_name)
            independent_permissions[resource_id] = typing.cast(Permissions, permissions_row.permissions)
        for resource_id in resource_ids:
            if resource_id not in independent_permissions:
                self._check_resource_exists(resource_id)
                independent_permissions[resource_id] = Permissions.NONE
        return independent_permissions

    def _set_user_independent_permissions(
            self,
            resource_id: int,
            permissions: Permissions,
            table: typing.Any
    ) -> None:
        """
        Set the permissions for a specific resource in a user independent table.

        :param resource_id: the ID of an existing resource
        :param permissions: the user independent permissions
        :param table: the permissions table
        """
        self._check_resource_exists(resource_id)

        if table is None:
            return

        if permissions == Permissions.NONE:
            table.query.filter_by(**{self._resource_id_name: resource_id}).delete()
        else:
            all_user_permissions = table.query.filter_by(**{self._resource_id_name: resource_id}).first()
            if all_user_permissions is None:
                all_user_permissions = table(permissions=permissions, **{self._resource_id_name: resource_id})
            else:
                all_user_permissions.permissions = permissions
            db.session.add(all_user_permissions)
        db.session.commit()

    def get_permissions_for_all_users(
            self,
            resource_id: int
    ) -> Permissions:
        """
        Return the permissions all users have for a specific resource.

        :param resource_id: the ID of an existing resource
        :return: the permissions for all users
        """
        return self._get_user_independent_permissions(resource_id, self._all_user_permissions_table)

    def get_permissions_for_all_users_for_multiple_resources(
            self,
            resource_ids: typing.Sequence[int]
    ) -> typing.Mapping[int, Permissions]:
        """
        Return the permissions all users have for a list of resources.

        :param resource_ids: the list of IDs of existing resources
        :return: the permissions for all users
        """
        return self._get_user_independent_permissions_for_multiple_resources(resource_ids, self._all_user_permissions_table)

    def set_permissions_for_all_users(
            self,
            resource_id: int,
            permissions: Permissions
    ) -> None:
        """
        Set the permissions all users have for a specific resource.

        :param resource_id: the ID of an existing resource
        :param permissions: the permissions for all users
        """
        self._set_user_independent_permissions(resource_id, permissions, self._all_user_permissions_table)

    def get_permissions_for_anonymous_users(
            self,
            resource_id: int
    ) -> Permissions:
        """
        Return the permissions anonymous users have for a specific resource.

        :param resource_id: the ID of an existing resource
        :return: the permissions for anonymous users
        """
        return self._get_user_independent_permissions(resource_id, self._anonymous_user_permissions_table)

    def get_permissions_for_anonymous_users_for_multiple_resources(
            self,
            resource_ids: typing.Sequence[int]
    ) -> typing.Mapping[int, Permissions]:
        """
        Return the permissions anonymous users have for a list of resources.

        :param resource_ids: the list of IDs of existing resources
        :return: the permissions for anonymous users
        """
        return self._get_user_independent_permissions_for_multiple_resources(resource_ids, self._anonymous_user_permissions_table)

    def set_permissions_for_anonymous_users(
            self,
            resource_id: int,
            permissions: Permissions
    ) -> None:
        """
        Set the permissions all users have for a specific resource.

        :param resource_id: the ID of an existing resource
        :param permissions: the permissions for anonymous users
        """
        self._set_user_independent_permissions(resource_id, permissions, self._anonymous_user_permissions_table)

    def get_permissions_for_users(
            self,
            resource_id: int,
            *,
            include_all_users: bool = False,
            include_groups: bool = False,
            include_projects: bool = False,
            include_admin_permissions: bool = False,
            limit_readonly_users: bool = False,
            additional_permissions: typing.Optional[typing.Dict[int, Permissions]] = None
    ) -> typing.Dict[int, Permissions]:
        """
        Get permissions for users for a specific resource.

        This does not consider modifications to user permissions other than
        those explicitly included using the function parameters.

        :param resource_id: the ID of an existing resource
        :param include_all_users: whether permissions for all users should be included
        :param include_groups: whether groups that the users are members of should be included
        :param include_projects: whether projects that the users are members of should be included
        :param include_admin_permissions: whether admin permissions should be included
        :param limit_readonly_users: whether readonly users should be limited to READ permissions
        :param additional_permissions: additional permissions to assume
        :return: a dict mapping users IDs to permissions
        """
        self._check_resource_exists(resource_id)

        permissions_for_users = {}
        if additional_permissions is not None:
            permissions_for_users = additional_permissions.copy()

        for user_permissions in self._user_permissions_table.query.filter_by(**{self._resource_id_name: resource_id}).all():
            permissions_for_users[user_permissions.user_id] = max(permissions_for_users.get(user_permissions.user_id, Permissions.NONE), user_permissions.permissions)

        if include_all_users:
            permissions_for_all_users = self.get_permissions_for_all_users(resource_id=resource_id)
            if permissions_for_all_users != Permissions.NONE:
                for user in users.get_users():
                    permissions_for_users[user.id] = max(permissions_for_users.get(user.id, Permissions.NONE), permissions_for_all_users)

        if include_groups:
            permissions_for_groups = self.get_permissions_for_groups(resource_id=resource_id, include_projects=include_projects)
            for group_id, group_permissions in permissions_for_groups.items():
                for user_id in groups.get_group_member_ids(group_id):
                    permissions_for_users[user_id] = max(permissions_for_users.get(user_id, Permissions.NONE), group_permissions)

        if include_projects:
            permissions_for_projects = self.get_permissions_for_projects(resource_id=resource_id)
            for project_id, project_permissions in permissions_for_projects.items():
                for user_id, member_permissions in projects.get_project_member_user_ids_and_permissions(project_id, include_groups=include_groups).items():
                    permissions_for_users[user_id] = max(permissions_for_users.get(user_id, Permissions.NONE), min(member_permissions, project_permissions))

        if include_admin_permissions:
            for user in users.get_administrators():
                if not user.has_admin_permissions:
                    # skip admins who do not use admin permissions
                    continue
                permissions_for_users[user.id] = Permissions.GRANT

        if limit_readonly_users:
            for user in users.get_users():
                if user.is_readonly and user.id in permissions_for_users:
                    permissions_for_users[user.id] = min(Permissions.READ, permissions_for_users[user.id])

        return {
            user_id: permissions
            for user_id, permissions in permissions_for_users.items()
            if permissions != Permissions.NONE
        }

    def set_permissions_for_user(
            self,
            resource_id: int,
            user_id: int,
            permissions: Permissions
    ) -> None:
        """
        Set the permissions for a user for a specific resource.

        Clear the permissions if called with Permissions.NONE

        :param resource_id: the ID of an existing resource
        :param user_id: the ID of an existing user
        :param permissions: the new permissions
        """
        self._check_resource_exists(resource_id)
        # ensure that the user can be found
        users.check_user_exists(user_id)

        if permissions == Permissions.NONE:
            self._user_permissions_table.query.filter_by(user_id=user_id, **{self._resource_id_name: resource_id}).delete()
        else:
            user_permissions = self._user_permissions_table.query.filter_by(user_id=user_id, **{self._resource_id_name: resource_id}).first()
            if user_permissions is None:
                user_permissions = self._user_permissions_table(user_id=user_id, permissions=permissions, **{self._resource_id_name: resource_id})
            else:
                user_permissions.permissions = permissions
            db.session.add(user_permissions)
        db.session.commit()

    def get_permissions_for_groups(
            self,
            resource_id: int,
            *,
            include_projects: bool = False
    ) -> typing.Dict[int, Permissions]:
        """
        Get permissions for groups for a specific resource.

        This does not consider modifications to user permissions other than
        those explicitly included using the function parameters.

        :param resource_id: the ID of an existing resource
        :param include_projects: whether projects that the groups are members of should be included
        :return: a dict mapping group IDs to permissions
        """
        self._check_resource_exists(resource_id)

        permissions_for_groups = {}
        for group_permissions in self._group_permissions_table.query.filter_by(**{self._resource_id_name: resource_id}).all():
            permissions_for_groups[group_permissions.group_id] = group_permissions.permissions

        if include_projects:
            permissions_for_projects = self.get_permissions_for_projects(resource_id=resource_id)
            for project_id, project_permissions in permissions_for_projects.items():
                for group_id, member_permissions in projects.get_project_member_group_ids_and_permissions(project_id).items():
                    permissions_for_groups[group_id] = max(permissions_for_groups.get(group_id, Permissions.NONE), min(member_permissions, project_permissions))
        return {
            group_id: permissions
            for group_id, permissions in permissions_for_groups.items()
            if permissions != Permissions.NONE
        }

    def set_permissions_for_group(
            self,
            resource_id: int,
            group_id: int,
            permissions: Permissions
    ) -> None:
        """
        Set the permissions for a group for a specific resource.

        Clear the permissions if called with Permissions.NONE.

        :param resource_id: the ID of an existing resource
        :param group_id: the ID of an existing group
        :param permissions: the new permissions
        """
        self._check_resource_exists(resource_id)
        # ensure that the group can be found
        groups.get_group(group_id)

        if permissions == Permissions.NONE:
            self._group_permissions_table.query.filter_by(group_id=group_id, **{self._resource_id_name: resource_id}).delete()
        else:
            group_permissions = self._group_permissions_table.query.filter_by(group_id=group_id, **{self._resource_id_name: resource_id}).first()
            if group_permissions is None:
                group_permissions = self._group_permissions_table(group_id=group_id, permissions=permissions, **{self._resource_id_name: resource_id})
            else:
                group_permissions.permissions = permissions
            db.session.add(group_permissions)
        db.session.commit()

    def get_permissions_for_projects(
            self,
            resource_id: int
    ) -> typing.Dict[int, Permissions]:
        """
        Get permissions for projects for a specific resource.

        :param resource_id: the ID of an existing resource
        :return: a dict mapping project IDs to permissions
        """
        self._check_resource_exists(resource_id)

        permissions_for_projects = {}
        for project_permissions in self._project_permissions_table.query.filter_by(**{self._resource_id_name: resource_id}).all():
            permissions_for_projects[project_permissions.project_id] = project_permissions.permissions
        return {
            project_id: permissions
            for project_id, permissions in permissions_for_projects.items()
            if permissions != Permissions.NONE
        }

    def set_permissions_for_project(
            self,
            resource_id: int,
            project_id: int,
            permissions: Permissions
    ) -> None:
        """
        Set the permissions for a project for a specific resource.

        Clear the permissions if called with Permissions.NONE.

        :param resource_id: the ID of an existing resource
        :param project_id: the ID of an existing project
        :param permissions: the new permissions
        """
        self._check_resource_exists(resource_id)
        # ensure that the project can be found
        projects.get_project(project_id)

        if permissions == Permissions.NONE:
            self._project_permissions_table.query.filter_by(project_id=project_id, **{self._resource_id_name: resource_id}).delete()
        else:
            project_permissions = self._project_permissions_table.query.filter_by(project_id=project_id, **{self._resource_id_name: resource_id}).first()
            if project_permissions is None:
                project_permissions = self._project_permissions_table(project_id=project_id, permissions=permissions, **{self._resource_id_name: resource_id})
            else:
                project_permissions.permissions = permissions
            db.session.add(project_permissions)
        db.session.commit()

    def get_permissions_for_user(
            self,
            resource_id: int,
            user_id: typing.Optional[int],
            *,
            include_all_users: bool = False,
            include_anonymous_users: bool = False,
            include_groups: bool = False,
            include_projects: bool = False,
            include_admin_permissions: bool = False,
            limit_readonly_users: bool = False,
            additional_permissions: Permissions = Permissions.NONE,
            max_permissions: Permissions = Permissions.GRANT
    ) -> Permissions:
        """
        Get combined permissions for a specific user for a specific resource.

        This does not consider instrument responsible users or other resource
        type dependent modifications to user permissions. Depending on the
        given arguments, it will however consider transitive permissions,
        admin permissions and readonly users.

        :param resource_id: the ID of an existing resource
        :param user_id: the ID of an existing user
        :param include_all_users: whether permissions for all users should be included
        :param include_anonymous_users: whether permissions for anonymous users should be included
        :param include_groups: whether groups that the users are members of should be included
        :param include_projects: whether projects that the users are members of should be included
        :param include_admin_permissions: whether admin permissions should be included
        :param limit_readonly_users: whether readonly users should be limited to READ permissions
        :param additional_permissions: additional permissions to assume for this user
        :param max_permissions: the maximum permissions to return
        :return: the combined permissions for the given user
        """
        return self.get_permissions_for_user_for_multiple_resources(
            resource_ids=[resource_id],
            user_id=user_id,
            include_all_users=include_all_users,
            include_anonymous_users=include_anonymous_users,
            include_groups=include_groups,
            include_projects=include_projects,
            include_admin_permissions=include_admin_permissions,
            limit_readonly_users=limit_readonly_users,
            additional_permissions={
                resource_id: additional_permissions
            },
            max_permissions=max_permissions
        )[resource_id]

    def get_permissions_for_user_for_multiple_resources(
            self,
            resource_ids: typing.Sequence[int],
            user_id: typing.Optional[int],
            *,
            include_all_users: bool = False,
            include_anonymous_users: bool = False,
            include_groups: bool = False,
            include_projects: bool = False,
            include_admin_permissions: bool = False,
            limit_readonly_users: bool = False,
            additional_permissions: typing.Optional[typing.Mapping[int, Permissions]] = None,
            max_permissions: Permissions = Permissions.GRANT
    ) -> typing.Mapping[int, Permissions]:
        """
        Get combined permissions for a specific user for multiple resources.

        This does not consider instrument responsible users or other resource
        type dependent modifications to user permissions. Depending on the
        given arguments, it will however consider transitive permissions,
        admin permissions and readonly users.

        :param resource_ids: a list of IDs of existing resources
        :param user_id: the ID of an existing user
        :param include_all_users: whether permissions for all users should be included
        :param include_anonymous_users: whether permissions for anonymous users should be included
        :param include_groups: whether groups that the users are members of should be included
        :param include_projects: whether projects that the users are members of should be included
        :param include_admin_permissions: whether admin permissions should be included
        :param limit_readonly_users: whether readonly users should be limited to READ permissions
        :param additional_permissions: additional permissions to assume for this user, or None
        :param max_permissions: the maximum permissions to return
        :return: a dict mapping the resource IDs to the combined permissions
            for the given user
        """
        if not resource_ids:
            return {}

        for resource_id in resource_ids:
            self._check_resource_exists(resource_id)

        permissions = {
            resource_id: additional_permissions.get(resource_id, Permissions.NONE) if additional_permissions else Permissions.NONE
            for resource_id in resource_ids
        }

        if include_anonymous_users:
            anonymous_user_permissions = self.get_permissions_for_anonymous_users_for_multiple_resources(
                resource_ids=[
                    resource_id
                    for resource_id in resource_ids
                    if max_permissions not in permissions[resource_id]
                ]
            )
            for resource_id in anonymous_user_permissions:
                permissions[resource_id] = max(permissions[resource_id], anonymous_user_permissions[resource_id])

        if user_id is None:
            return permissions

        # ensure that the user can be found
        user = users.get_user(user_id)

        # readonly users may never have more than READ permissions
        if user.is_readonly and limit_readonly_users:
            max_permissions = min(Permissions.READ, max_permissions)

        if include_admin_permissions and user.has_admin_permissions:
            for resource_id in resource_ids:
                if max_permissions not in permissions[resource_id]:
                    # administrators have GRANT permissions if they use admin permissions
                    permissions[resource_id] = max(permissions[resource_id], Permissions.GRANT)

        if include_all_users:
            all_user_permissions = self.get_permissions_for_all_users_for_multiple_resources(
                resource_ids=[
                    resource_id
                    for resource_id in resource_ids
                    if max_permissions not in permissions[resource_id]
                ]
            )
            for resource_id in all_user_permissions:
                permissions[resource_id] = max(permissions[resource_id], all_user_permissions[resource_id])

        non_max_permissions_resource_ids = [
            resource_id
            for resource_id in resource_ids
            if max_permissions not in permissions[resource_id]
        ]
        if non_max_permissions_resource_ids:
            resource_id_column = getattr(self._user_permissions_table, self._resource_id_name)
            permissions_rows = self._user_permissions_table.query.filter_by(
                user_id=user_id
            ).filter(
                resource_id_column.in_(non_max_permissions_resource_ids)
            ).all()
            for permissions_row in permissions_rows:
                resource_id = getattr(permissions_row, self._resource_id_name)
                permissions[resource_id] = max(permissions[resource_id], permissions_row.permissions)

        if include_groups:
            non_max_permissions_resource_ids = [
                resource_id
                for resource_id in resource_ids
                if max_permissions not in permissions[resource_id]
            ]
            if non_max_permissions_resource_ids:
                group_ids = [
                    group.id
                    for group in groups.get_user_groups(user_id)
                ]
                if group_ids:
                    resource_id_column = getattr(self._group_permissions_table, self._resource_id_name)
                    permissions_rows = self._group_permissions_table.query.filter(
                        self._group_permissions_table.group_id.in_(group_ids),
                        resource_id_column.in_(non_max_permissions_resource_ids)
                    ).all()
                    for permissions_row in permissions_rows:
                        resource_id = getattr(permissions_row, self._resource_id_name)
                        permissions[resource_id] = max(permissions[resource_id], permissions_row.permissions)

        if include_projects:
            non_max_permissions_resource_ids = [
                resource_id
                for resource_id in resource_ids
                if max_permissions not in permissions[resource_id]
            ]
            if non_max_permissions_resource_ids:
                project_ids = [
                    project.id
                    for project in projects.get_user_projects(user_id, include_groups=include_groups)
                ]
                if project_ids:
                    user_project_permissions = {
                        project_id: projects.get_user_project_permissions(project_id=project_id, user_id=user_id, include_groups=include_groups)
                        for project_id in project_ids
                    }
                    resource_id_column = getattr(self._project_permissions_table, self._resource_id_name)
                    permissions_rows = self._project_permissions_table.query.filter(
                        self._project_permissions_table.project_id.in_(project_ids),
                        resource_id_column.in_(non_max_permissions_resource_ids)
                    ).all()
                    for permissions_row in permissions_rows:
                        resource_id = getattr(permissions_row, self._resource_id_name)
                        max_project_permissions = user_project_permissions[permissions_row.project_id]
                        permissions[resource_id] = max(permissions[resource_id], min(max_project_permissions, permissions_row.permissions))

        return {
            resource_id: min(permissions[resource_id], max_permissions)
            for resource_id in resource_ids
        }

    def copy_permissions(self, source_resource_id: int, target_resource_id: int) -> None:
        self._check_resource_exists(source_resource_id)
        self._check_resource_exists(target_resource_id)

        # clear current permissions for the target resource
        self._all_user_permissions_table.query.filter_by(**{self._resource_id_name: target_resource_id}).delete()
        self._user_permissions_table.query.filter_by(**{self._resource_id_name: target_resource_id}).delete()
        self._group_permissions_table.query.filter_by(**{self._resource_id_name: target_resource_id}).delete()
        self._project_permissions_table.query.filter_by(**{self._resource_id_name: target_resource_id}).delete()

        permissions_for_all_users = self.get_permissions_for_all_users(source_resource_id)
        self.set_permissions_for_all_users(target_resource_id, permissions_for_all_users)
        permissions_for_users = self.get_permissions_for_users(source_resource_id)
        for user_id, permissions in permissions_for_users.items():
            self.set_permissions_for_user(target_resource_id, user_id, permissions)
        permissions_for_groups = self.get_permissions_for_groups(source_resource_id)
        for group_id, permissions in permissions_for_groups.items():
            self.set_permissions_for_group(target_resource_id, group_id, permissions)
        permissions_for_projects = self.get_permissions_for_projects(source_resource_id)
        for project_id, permissions in permissions_for_projects.items():
            self.set_permissions_for_project(target_resource_id, project_id, permissions)
